import pytest

from spearmint import Experiment, HypothesisTest
from spearmint.utils import generate_fake_observations


@pytest.fixture()
def binary_data():
    return generate_fake_observations(
        distribution="bernoulli",
        n_treatments=6,
        n_attributes=1,
        n_observations=6 * 1000,
        random_seed=123,
    )


def test_large_proportions_delta_expermiment(binary_data):
    exp = Experiment(binary_data)

    # run 'A/A' test
    test_aa = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        hypothesis="larger",
        inference_method="frequentist",
        variable_type="binary",
    )
    results_aa = exp.run_test(test_aa)

    results_aa.display()
    assert results_aa.test_statistic_name == "z"
    assert not results_aa.accept_hypothesis

    # run A/B test
    test_ab = HypothesisTest(
        metric="metric",
        control="A",
        variation="B",
        hypothesis="larger",
        inference_method="frequentist",
        variable_type="binary",
    )
    results_ab = exp.run_test(test_ab)
    results_ab.display()

    assert results_ab.test_statistic_name == "z"
    assert results_ab.accept_hypothesis


def test_proportions_delta_ab_unequal(binary_data):
    exp = Experiment(binary_data)

    # run A/B test
    test_ab = HypothesisTest(
        metric="metric",
        control="A",
        variation="F",
        hypothesis="unequal",
        inference_method="frequentist",
        variable_type="binary",
    )
    results_ab = exp.run_test(test_ab)

    results_ab.display()
    assert results_ab.test_statistic_name == "z"
    assert results_ab.accept_hypothesis


def test_proportions_delta_ab_larger(binary_data):
    exp = Experiment(binary_data)

    # run A/B test
    test_ab = HypothesisTest(
        metric="metric",
        control="A",
        variation="F",
        hypothesis="larger",
        inference_method="frequentist",
        variable_type="binary",
    )
    results_ab = exp.run_test(test_ab)
    results_ab.display()
    assert results_ab.accept_hypothesis


def test_proportions_delta_ab_smaller(binary_data):
    exp = Experiment(binary_data)

    # run A/B test
    test_ab = HypothesisTest(
        metric="metric",
        control="A",
        variation="F",
        hypothesis="smaller",
        inference_method="frequentist",
        variable_type="binary",
    )
    results_ab = exp.run_test(test_ab)
    results_ab.display()
    assert not results_ab.accept_hypothesis


def test_proportions_delta_aa(binary_data):
    exp = Experiment(binary_data)

    # run A/A test
    test_aa = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        hypothesis="larger",
        inference_method="frequentist",
        variable_type="binary",
    )
    results_aa = exp.run_test(test_aa)
    results_aa.display()
    assert not results_aa.accept_hypothesis


def test_proportions_delta_default(binary_data):
    exp = Experiment(binary_data)

    # run A/B test
    test_ab = HypothesisTest(metric="metric", control="A", variation="F")
    results_ab = exp.run_test(test_ab)
    results_ab.display()
    assert results_ab.accept_hypothesis
    assert test_ab.inference_method == "frequentist"
    assert test_ab.variable_type == "binary"
    assert results_ab.accept_hypothesis



import numpy as np
import pytest

from spearmint import stats

TEST_SAMPLES_NAME = "TestSamples"
TEST_CONTINUOUS_SAMPLE_VARIANCE = 1.0
TEST_BINARY_SAMPLE_VARIANCE = 0.5**2
TEST_COUNT_SAMPLE_RATE = 10


@pytest.fixture
def continuous_test_samples():
    np.random.seed(123)
    return stats.Samples(
        np.random.randn(1000) * TEST_CONTINUOUS_SAMPLE_VARIANCE**0.5,
        name=TEST_SAMPLES_NAME,
    )


@pytest.fixture
def binary_test_samples():
    np.random.seed(123)
    return stats.Samples(
        np.random.rand(1000),
        name=TEST_SAMPLES_NAME,
    )


@pytest.fixture
def count_test_samples():
    np.random.seed(123)
    return stats.Samples(
        np.random.poisson(TEST_COUNT_SAMPLE_RATE, 1000),
        name=TEST_SAMPLES_NAME,
    )


def test_bonferroni_correction():
    p_vals_orig = [0.05, 0.05]
    alpha_orig = 0.05
    corrected = stats.bonferroni_correction(alpha_orig, p_vals_orig)
    assert corrected < alpha_orig


def test_sidak_correction():
    p_vals_orig = [0.05, 0.05]
    alpha_orig = 0.05
    corrected = stats.sidak_correction(alpha_orig, p_vals_orig)
    assert corrected < alpha_orig


def test_fdr_bh_correction():
    p_vals_orig = [0.5, 0.5]
    fdr_orig = 0.05
    corrected = stats.fdr_bh_correction(fdr_orig, p_vals_orig)
    assert corrected < fdr_orig

    p_vals_orig = [0.05, 0.05]
    corrected = stats.fdr_bh_correction(fdr_orig, p_vals_orig)
    assert corrected == fdr_orig


def test_multiple_comparison_correction():
    p_values = np.arange(0.001, 0.1, 0.01)
    mc = stats.MultipleComparisonCorrection(p_values, method="b")

    assert mc.alpha_corrected < mc.alpha_orig
    assert mc.accept_hypothesis[0]
    assert not mc.accept_hypothesis[-1]

    with pytest.raises(ValueError):
        stats.MultipleComparisonCorrection(p_values, method="unknown")


def test_estimate_experiment_sample_sizes_z():
    prob_control = 0.49
    std_control = (prob_control * (1 - prob_control)) ** 0.5  # Binomial std
    prob_variation = std_variation = 0.50
    delta = prob_variation - prob_control
    assert stats.estimate_experiment_sample_sizes(
        delta=delta, statistic="z", std_control=std_control, std_variation=std_variation
    ) == (39236, 39236)


def test_estimate_experiment_sample_sizes_t():
    prob_control = 0.49
    std_control = (prob_control * (1 - prob_control)) ** 0.5  # Binomial std
    prob_variation = std_variation = 0.50
    delta = prob_variation - prob_control
    assert stats.estimate_experiment_sample_sizes(
        delta=delta, statistic="t", std_control=std_control, std_variation=std_variation
    ) == (39237, 39237)


def test_estimate_experiment_sample_sizes_ratio():
    # Replicate Example 1 from Gu et al, 2008
    R = 4  # ratio under alternative hypothesis
    control_rate = 0.0005
    variation_rate = R * control_rate
    delta = variation_rate - control_rate
    assert stats.estimate_experiment_sample_sizes(
        delta,
        statistic="rates_ratio",
        control_rate=control_rate,
        alpha=0.05,
        power=0.9,
        control_exposure_time=2.0,
        sample_size_ratio=0.5,
    ) == (8590, 4295)


def test_estimate_experiment_sample_sizes_unknown():
    with pytest.raises(ValueError):
        stats.estimate_experiment_sample_sizes(delta=None, statistic="unknown")


def test_cohens_d_sample_size_unknown_statistic():
    with pytest.raises(ValueError):
        stats.cohens_d_sample_size(
            delta=1, alpha=0.05, power=0.8, std_control=1, statistic="unknown"
        )


def test_highest_density_interval():
    samples = np.random.randn(1000)
    hdi = stats.highest_density_interval(samples, credible_mass=0.9)
    assert hdi[0] >= -1.96  # test lower HDI for 95% hdi
    assert hdi[1] <= 1.96  # test upper HDI for 95% hdi
    with pytest.raises(ValueError):
        # Not enough samples
        stats.highest_density_interval([1], 1.0)


def test_empirical_cdf_class(continuous_test_samples):
    observations = continuous_test_samples._raw_observations
    ecdf = stats.EmpiricalCdf(observations)
    assert ecdf.samples_cdf[-1] == 1.0

    # __call__
    assert np.all(ecdf([-100, 100]) == np.array([0.0, 1.0]))

    # evaluate
    assert np.all(ecdf.evaluate([-100, 100]) == np.array([0.0, 1.0]))
    assert len(ecdf.evaluate()) == len(observations)


def test_samples(continuous_test_samples):
    assert continuous_test_samples.mean <= 0.01
    assert continuous_test_samples.std - 1 <= 0.01
    assert not np.all(
        continuous_test_samples.permute() == continuous_test_samples.permute()
    )
    assert np.all(
        continuous_test_samples.sort()
        == sorted(continuous_test_samples._raw_observations)
    )
    continuous_test_samples.summary  # print the table
    assert len(continuous_test_samples._summary_table._print_history) > 0


def test_samples_summary_table(continuous_test_samples):
    samples_table = stats.SamplesSummaryTable(samples=continuous_test_samples)
    samples_table.print()
    printed_lines = samples_table._print_history[-1].split("\n")

    assert "Samples Summary" in printed_lines[0]
    assert TEST_SAMPLES_NAME in printed_lines[2]
    assert "Samples" in printed_lines[4]
    assert "Mean" in printed_lines[5]
    assert "Standard Error" in printed_lines[6]
    assert "Variance" in printed_lines[7]


def test_continuous_samples_comparison_table(continuous_test_samples):
    control_samples = continuous_test_samples
    variation_samples = continuous_test_samples

    # Single variation instance
    samples_table = stats.SamplesComparisonTable(
        control_samples=control_samples, variation_samples=variation_samples
    )
    samples_table.print()
    printed_lines = samples_table._print_history[-1].split("\n")

    assert "Samples Comparison" in printed_lines[0]
    assert TEST_SAMPLES_NAME in printed_lines[2]
    assert "Mean" in printed_lines[5]
    assert "Standard Error" in printed_lines[6]
    assert "Variance" in printed_lines[7]
    assert "Delta" in printed_lines[8]

    # Multiple variation instance
    samples_table = stats.SamplesComparisonTable(
        control_samples=control_samples,
        variation_samples=[variation_samples, variation_samples],
    )
    samples_table.print()
    printed = samples_table._print_history[-1]
    assert "Samples Comparison" in printed
    assert TEST_SAMPLES_NAME in printed


def test_mean_comparison(continuous_test_samples):
    # Same samples, so
    means_comparison = stats.MeanComparison(
        samples_a=continuous_test_samples, samples_b=continuous_test_samples
    )
    means_comparison.comparison  # test printed summary table
    assert len(means_comparison._comparison_table._print_history) > 0
    assert means_comparison.delta == 0
    assert means_comparison.alpha == stats.DEFAULT_ALPHA
    assert means_comparison.test_direction == stats.DEFAULT_HYPOTHESIS
    assert np.isclose(
        means_comparison.pooled_variance, TEST_CONTINUOUS_SAMPLE_VARIANCE, atol=0.01
    )
    assert means_comparison.delta_relative == 0
    assert means_comparison.effect_size == 0
    # overlapping distributions, power should be similar to false positive rate
    assert np.isclose(means_comparison.power, means_comparison.alpha, atol=0.001)
    t_test_stats = means_comparison.t_test_stats
    assert t_test_stats["statistic_name"] == "t"
    assert t_test_stats["statistic_value"] == 0
    assert t_test_stats["p_value"] == 0.5
    assert t_test_stats["alpha"] == stats.DEFAULT_ALPHA
    assert np.isclose(t_test_stats["power"], t_test_stats["alpha"], atol=0.001)

    assert (
        t_test_stats["degrees_freedom"]
        == (continuous_test_samples.nobs + continuous_test_samples.nobs) - 2
    )


def test_proportion_comparison(binary_test_samples):
    # Same samples
    proportion_comparison = stats.ProportionComparison(
        samples_a=binary_test_samples, samples_b=binary_test_samples
    )
    assert proportion_comparison.delta == 0
    assert np.isclose(
        proportion_comparison.pooled_variance, TEST_BINARY_SAMPLE_VARIANCE, atol=0.01
    )
    assert proportion_comparison.delta_relative == 0
    assert proportion_comparison.effect_size == 0
    # overlapping distributions, power should be similar to false positive rate
    assert np.isclose(
        proportion_comparison.power, proportion_comparison.alpha, atol=0.001
    )

    z_test_stats = proportion_comparison.z_test_stats
    assert z_test_stats["statistic_name"] == "z"
    assert z_test_stats["statistic_value"] == 0
    assert z_test_stats["p_value"] == 0.5
    assert z_test_stats["alpha"] == stats.DEFAULT_ALPHA
    assert np.isclose(z_test_stats["power"], z_test_stats["alpha"], atol=0.001)


def test_rates_comparison(count_test_samples):
    # Same samples
    rate_comparison = stats.RateComparison(
        samples_a=count_test_samples, samples_b=count_test_samples
    )

    assert rate_comparison.delta == 1.0
    assert rate_comparison.delta_relative == 1.0
    assert np.isclose(rate_comparison.pooled_variance, TEST_COUNT_SAMPLE_RATE, atol=0.1)
    assert rate_comparison.effect_size == 0
    # overlapping distributions, power should be similar to false positive rate
    assert np.isclose(rate_comparison.power, rate_comparison.alpha, atol=0.001)

    rates_test_stats = rate_comparison.rates_test_stats
    assert rates_test_stats["statistic_name"] == "W"
    assert rates_test_stats["statistic_value"] == 0
    assert rates_test_stats["p_value"] == 0.5
    assert rates_test_stats["alpha"] == stats.DEFAULT_ALPHA
    assert np.isclose(rates_test_stats["power"], rates_test_stats["alpha"], atol=0.001)


def test_bootstrap_comparison(continuous_test_samples):
    # Same samples
    bs_comparison = stats.BootstrapStatisticComparison(
        samples_a=continuous_test_samples,
        samples_b=continuous_test_samples,
        n_bootstraps=1000,
    )

    assert np.isclose(bs_comparison.delta, 0.0, atol=0.001)
    assert np.isclose(bs_comparison.delta_relative, 0.0, atol=0.05)
    assert np.isclose(
        bs_comparison.pooled_variance, TEST_CONTINUOUS_SAMPLE_VARIANCE, atol=0.1
    )
    assert np.isclose(bs_comparison.effect_size, 0.0, atol=0.001)

    # overlapping distributions, power should be similar to false positive rate
    assert np.isclose(bs_comparison.power, bs_comparison.alpha, atol=0.05)

    bootstrap_test_stats = bs_comparison.bootstrap_test_stats
    assert bootstrap_test_stats["statistic_function_name"] == "mean"
    assert bootstrap_test_stats["statistic_name"] == "bootstrap_mean"
    assert np.isclose(bootstrap_test_stats["statistic_value"], 0, atol=0.001)
    assert np.isclose(bootstrap_test_stats["p_value"], 0.5, atol=0.05)
    assert bootstrap_test_stats["alpha"] == stats.DEFAULT_ALPHA
    assert np.isclose(
        bootstrap_test_stats["power"], bootstrap_test_stats["alpha"], atol=0.01
    )



from copy import deepcopy
from datetime import datetime

from spearmint.config import (
    DEFAULT_ALPHA,
    DEFAULT_INFERENCE_METHOD,
    DEFAULT_HYPOTHESIS,
    DEFAULT_METRIC_NAME,
    DEFAULT_TREATMENT_NAME,
)
from spearmint.inference import InferenceResults, get_inference_procedure
from spearmint.stats import MultipleComparisonCorrection, Samples
from spearmint.typing import Callable, DataFrame, FilePath, Optional, List, Union
from spearmint.utils import ensure_dataframe, infer_variable_type


class CohortFilter:
    """
    Filtering interface for selecting cohorts of a dataframe. Powered pandas
    `query` method.

    Parameters
    ----------
    treatment_column : str
        The column associated with various treatments
    treatment_name : str
        When applied, selects out rows with `treatment_column` == `treatment_name`
    """

    def __init__(self, treatment_column: str, treatment_name: str):
        self.treatment_column = treatment_column
        self.treatment_name = treatment_name

    def apply(self, data: DataFrame) -> DataFrame:
        """
        Parameters
        ----------
        data : DataFrame

        Returns
        -------
        cohort_data : DataFrame
            cohort data, filtered from `data`
        """
        return data.query(f"`{self.treatment_column}` == '{self.treatment_name}'")


class SegmentFilter:
    """
    Simple filtering interface for selecting subsets of a dataframe. Powered
    pandas `query` interface.
    """

    def __init__(self, segment_query: Optional[str] = None):
        self.segment_query = segment_query

    def apply(self, data: DataFrame) -> DataFrame:
        if self.segment_query is not None:
            return data.query(self.segment_query)

        return data


class CustomMetric:
    """
    Metric definition that combines one or more measure columns in dataframe.

    Parameters
    ----------
    metric_function: Callable
        the function that defines an operation performed on each row of the
        dataframe, resulting in a derived value.
    """

    def __init__(self, metric_function: Callable):
        self.metric_function = metric_function

    def apply(self, data: DataFrame) -> DataFrame:
        return data.apply(self.metric_function, axis=1)


class HypothesisTest:
    """
    Executes directives for running inference a procudure to compare the central
    tendencies of two groups of random samples.

    Parameters
    ----------

    """

    def __init__(
        self,
        control: str,
        variation: str,
        metric: Optional[Union[str, CustomMetric]] = DEFAULT_METRIC_NAME,
        treatment: Optional[Union[str, CustomMetric]] = DEFAULT_TREATMENT_NAME,
        inference_method: str = DEFAULT_INFERENCE_METHOD,
        hypothesis: Optional[str] = DEFAULT_HYPOTHESIS,
        segmentation: Optional[Union[str, List[str]]] = None,
        variable_type: Optional[str] = None,
        **inference_procedure_params,
    ):
        """
        Parameters
        ----------
        control : str
            The name of the control treatment
        variation : str
            The name of the experimental treatment
        metric: str | CustomMetric (optional)
            a performance indicator over which statistical analyses
            will be performed. Each can either be a measurment in the experiment's
            dataset (i.e. a column name), or an instance of a CustomMetric. If
            None provided, we use the setting in
            spearmint.cfg::hypothesis_test.default_metric_name
        inference_method : str (optional)
            The name of the inference method used to perform the hypothesis test.
            Can be one of "frequentist", "bayesian", "bootstrap". If None provided,
            we use the setting in
            spearmint.cfg::hypothesis_test.default_inference_method
        hypothesis: str (optional)
            One of "larger", "smaller", "unqual". The hypothesis to test when
            comparing the `control` and `variation` groups. If
            None provided, we use the setting in
            spearmint.cfg::hypothesis_test.default_hypothesis
        treatment: str (optional)
            The name of the column in the experiement observations associated
            with treatment naems. If None provided, we use the setting in
            spearmint.cfg::hypothesis_test.default_treatment_name
        segmentation : str | List[str] (optional)
            Defines a list of logical filter operations that follows the conventions
            used by Panda's dataframe query api and segments the treatments into subgroups.
            If a list provided, all operations are combined using logical AND
        variable_type : optional str, None
            One of 'continuous', 'binary', 'counts'. Explicitly declares the variable
            type. If None provided, we infer the variable type from the values
            of `metric`.
        **inference_procedure_params : dict
            Any additional parameters used to initialize the inference procedure

        """
        self.metric = metric
        self.treatment = treatment
        self.control = control
        self.variation = variation
        self.inference_method = (
            inference_method.lower().replace("-", "_").replace(" ", "_")
        )
        self.variable_type = variable_type
        self.hypothesis = hypothesis

        # TODO: add back in segmentation to InferenceProcedure
        if isinstance(segmentation, list):
            segmentation = " & ".join(segmentation)
        self.segmentation = segmentation

        if isinstance(self.metric, CustomMetric):
            self.metric_column = self.metric.metric_function.__name__
        else:
            self.metric_column = str(metric)

        self.inference_procedure_params = inference_procedure_params

    def _add_custom_metric_column(self, _data):
        data = _data.copy()
        self.metric_name = self.metric.metric_function.__name__
        data[self.metric_name] = self.metric.apply(data)
        return data

    def filter_variations(self, data: DataFrame, variation_name: str) -> DataFrame:
        """
        Helper function used to pull out the data series observations for a
        variation.
        """
        data = ensure_dataframe(data)
        if isinstance(self.metric, CustomMetric):
            data = self._add_custom_metric_column(data)
        else:
            self.metric_name = self.metric

        cohort_filter = CohortFilter(self.treatment, variation_name)  # type: ignore
        return cohort_filter.apply(data)

    def filter_segments(self, data: DataFrame) -> DataFrame:
        """
        Helper function used to filter observations for all segments.
        """
        data = ensure_dataframe(data)
        segment_filter = SegmentFilter(self.segmentation)
        return segment_filter.apply(data)

    def filter_metrics(self, data: DataFrame) -> DataFrame:
        """
        Helper function used to filter out observations that have invalid metric
        values.
        """
        return data[data.notna()[self.metric_name]][self.metric_name]

    def run(
        self,
        control_samples: Samples,
        variation_samples: Samples,
        alpha: float = DEFAULT_ALPHA,
    ) -> InferenceResults:
        """
        Run the statistical test inference procedure comparing two groups of
        samples.

        Parameters
        ----------
        control_samples : an instance of `Samples` class
            The samples for the control treatment group
        variation_samples : an instance of `Samples` class
            The samples for the experimental treatment group
        alpha : float in (0, 1) (optional)
            Is either:
                -   the 'significance level' for frequentist tests
                -   the one minus the credible mass of the posterior over differences
                    between the two groups.
            If None provided, we assume an alpha = 0.05
        inference_kwargs : dict
            Any additional keyword args to be provided to the inference procedure

        Returns
        -------
        results : InferenceResults
            The object class holding the summary of the experiment

        """
        self.variable_type = (
            self.variable_type
            if self.variable_type is not None
            else infer_variable_type(control_samples.data)
        )

        self.inference_procedure = get_inference_procedure(
            variable_type=self.variable_type,
            inference_method=self.inference_method,
            metric_name=self.metric_column,
            hypothesis=self.hypothesis,
            **self.inference_procedure_params,
        )

        # Update alphas
        self.alpha = alpha
        self.inference_procedure.alpha = alpha

        results = self.inference_procedure.run(
            control_samples,
            variation_samples,
        )

        results.metric_name = self.metric_column
        # results.segmentation = self.segmentation
        return results

    def copy(self, **update_kwargs) -> "HypothesisTest":
        """
        Make a copy of the current hypothesis test, substituting any provided
        values in `**update_kwargs`
        """
        copy = deepcopy(self)
        inference_kwargs = update_kwargs.get("infer_kwargs", {})
        for k, v in update_kwargs.items():
            setattr(copy, k, v)

        copy.inference_procedure = get_inference_procedure(
            copy.variable_type, copy.inference_method, **inference_kwargs  # type: ignore  # `.variable_type` inherited, or determined during inference
        )
        return copy


class HypothesisTestGroup:
    """
    A group of simultaneous hypothesis test. Performs multiple correction.

    Parameters
    ----------
    tests : Iterable[HypothesisTest]
        The tests to run
    correction_method : str
        One of the following correction methods:
            'bonferroni', 'b' : one-step Bonferroni correction
            'sidak', 's' : one-step Sidak correction
            'fdr_bh', 'bh; : Benjamini/Hochberg (non-negative)
    """

    def __init__(self, tests: List[HypothesisTest], correction_method: str = "sidak"):
        self.tests = tests
        self.correction_method = correction_method


class GroupInferenceResults:
    """
    Store, display, visualize, and export the results of statistical test
    suite.
    """

    def __init__(
        self,
        test_group: HypothesisTestGroup,
        correction: MultipleComparisonCorrection,
        original_results: List[InferenceResults],
        corrected_results: List[InferenceResults],
    ) -> None:
        self.tests = test_group.tests
        self.ntests = len(self.tests)
        self.original_results = original_results
        self.corrected_results = corrected_results
        self.correction = correction
        self.run_at: Optional[datetime] = None

    def display(self) -> None:
        for ii, res in enumerate(self.corrected_results):
            print("-" * 60 + f"\nTest {ii + 1} of {self.ntests}")
            res.display()

    def visualize(self, outfile: Optional[FilePath] = None, *args, **kwargs):
        for _, res in enumerate(self.corrected_results):
            res.visualize(outfile=outfile)



from abc import ABC, abstractmethod
from collections import OrderedDict
from enum import Enum
from datetime import datetime

from holoviews import Element

from spearmint.typing import FilePath, List, Callable, Optional, Any, Union
from spearmint.mixin import DataframeableMixin
from spearmint.stats import Samples, SamplesComparisonTable, DEFAULT_ALPHA
from spearmint.utils import process_warnings


class InferenceResultsDisplayError(Exception):
    """Raised when there is an error displaying inference procedure results"""

    pass


class InferenceResultsVisualizationError(Exception):
    """Raised when there is an error visualizaing inference procedure results"""

    pass


class InferenceResultsMissingError(Exception):
    """Raised when there are no results associated with an inference procedure."""

    pass


class VariableType(str, Enum):
    continuous = "continuous"
    binary = "binary"
    counts = "counts"


class InferenceMethod(str, Enum):
    frequentist = "frequentist"
    bayesian = "bayesian"
    bootstrap = "bootstrap"


class Hypothesis(str, Enum):
    larger = "larger"
    smaller = "smaller"
    unequal = "unequal"


class InferenceResults(DataframeableMixin):
    """
    Base class for displaying and saving statistical inference results

    Parameters
    ----------
    control : Samples
        the control samples
    variation : Samples
        the variation samples
    metric_name : str
        The name of the metric that is being compared
    delta: float
        the absolute difference between the variation and control sample means
    delta_relative: float
        the percent difference between the variation and control sample means
    hypothesis : str
        Human-readable message for interpreting the experiment
    alpha : float in (0, 1)
        The "significance" level, or type I error rate for the experiment
    accept_hypothesis : boolean
        Whether or not to accept the alternative hypothesis, based on `p_value`
        and `alpha` and any correction method.
    warnings : list[str]
        A list of any warning messages accumultaed during the test
    aux : dict
        Auxillary variables used for displaying or visualizing specific types
        of tests.
    """

    def __init__(
        self,
        control: Samples,
        variation: Samples,
        metric_name: str,
        delta: float,
        delta_relative: float,
        effect_size: float,
        hypothesis: Hypothesis,
        alpha: float,
        accept_hypothesis: bool,
        inference_method: InferenceMethod,
        variable_type: VariableType,
        visualization_function: Callable,
        comparison_type: Optional[str] = None,
        warnings: Optional[Union[str, List[str]]] = None,
        correction_method: Optional[str] = None,
        segmentation: Optional[str] = None,
        aux: dict = {},
    ):
        self.control = control
        self.variation = variation
        self.metric_name = metric_name
        self.delta = delta
        self.delta_relative = delta_relative
        self.effect_size = effect_size
        self.hypothesis = hypothesis
        self.variable_type = variable_type
        self.alpha = alpha
        self.accept_hypothesis = accept_hypothesis
        self.inference_method = inference_method
        self.comparison_type = comparison_type
        self.warnings = process_warnings(warnings)
        self.aux = aux
        self.visualization_function = visualization_function
        self.correction_method = correction_method
        self.segmentation = segmentation

        self.created_at: datetime = datetime.now()

        # These properties should be updated by the running inference proc.
        self.run_at: Optional[datetime] = None
        self.p_value: Optional[float] = None  # only Frequentist/Bootstrap tests

    def display(self) -> None:
        """Display the inference procedure results to the console"""
        try:
            self.summary
        except Exception as e:
            raise InferenceResultsDisplayError(e)

    def visualize(
        self, outfile: Optional[FilePath] = None, *args, **kwargs
    ) -> Element:  # pragma: no cover
        """Visualize the inference procedure results

        Parameters
        ----------
        outfile : FilePath, optional
            If provided, use to save the output visualization, by default None

        Returns
        -------
        Element : holoviews.Element
            A constructed visualization object

        Raises
        ------
        InferenceResultsVisualizationError
            If the `_render_visualization` is not implemented. Note: we do not
            require `_render_visualization` via abstractmethod, leaving the
            option for std-out-only reporting
        """
        try:
            return self.visualization_function(self, outfile, *args, **kwargs)

        except Exception as e:
            raise InferenceResultsVisualizationError(e)

    def _render_stats_table(self):
        raise NotImplementedError(
            "No implementation of `_render_stats_table` for ",
            f"class {self.__name__}, cannot execute `.summary`",
        )

    def to_csv(self, outfile: FilePath, delimiter=","):
        """
        Export result to delimiter-separated value file
        """
        results_df = self.to_dataframe()
        results_df.to_csv(outfile, sep=delimiter, encoding="utf8", index=False)

    @property
    def hypothesis_text(self):
        if self.hypothesis == "larger":
            return f"{self.variation.name} is larger"
        elif self.hypothesis == "smaller":
            return f"{self.variation.name} is smaller"
        elif self.hypothesis == "unequal":
            return f"{self.variation.name} != {self.control.name}"
        else:
            raise ValueError("Unknown hypothesis: {self.hypothesis}")

    @property
    def samples_comparison(self):
        if not hasattr(self, "_samples_comparison_table"):
            self._samples_comparison_table = SamplesComparisonTable(
                self.control, self.variation
            )
        self._samples_comparison_table.print()

    @property
    def stats(self):
        if not hasattr(self, "_stats_table"):
            self._stats_table = self._render_stats_table()
        self._stats_table.print()

    @property
    def summary(self):
        self.samples_comparison
        self.stats

    @property
    def _base_properties(self):
        """
        Properties that belong to all inference procedure results
        """
        return OrderedDict(
            [
                ("control_name", self.control.name),
                ("control_nobs", self.control.nobs),
                ("control_mean", self.control.mean),
                ("control_ci", self.control.confidence_interval(self.alpha)),
                ("control_var", self.control.var),
                ("variation_name", self.variation.name),
                ("variation_nobs", self.variation.nobs),
                ("variation_mean", self.variation.mean),
                ("variation_ci", self.variation.confidence_interval(self.alpha)),
                ("variation_var", self.variation.var),
                ("metric", self.metric_name),
                ("delta", self.delta),
                ("delta_relative", 100 * self.delta_relative),
                ("effect_size", self.effect_size),
                ("alpha", self.alpha),
                ("hypothesis", self.hypothesis),
                ("accept_hypothesis", self.accept_hypothesis),
                ("inference_method", self.inference_method),
                ("variable_type", self.variable_type),
                ("segmentation", self.segmentation),
                ("warnings", self.warnings),
            ]
        )

    @property
    def _specific_properties(self) -> OrderedDict:
        """
        Properties for specific inference procedures
        """
        return OrderedDict()

    def to_dict(self) -> OrderedDict[Any, Any]:
        _dict = self._base_properties
        _dict.update(self._specific_properties)
        assert isinstance(_dict, OrderedDict)
        return _dict

    @property
    def dict(self) -> OrderedDict:
        return self.to_dict()


class InferenceProcedure(ABC):
    """
    Base class for all inference procedures. Must implement the following
    private abstracmethods:

        -   `_run_inference()` : runs the inference procedure, updating any internal
            state needed for displaying and/or visualizing results
        -   `_make_results()` : generate `InferenceResults` based on the outcome
            of `_run_inference()`
    """

    def __init__(
        self,
        variable_type: VariableType = VariableType.continuous,
        inference_method: InferenceMethod = InferenceMethod.frequentist,
        hypothesis: Hypothesis = Hypothesis.larger,
        alpha: float = DEFAULT_ALPHA,
        metric_name: Optional[str] = None,
        **inference_procedure_params,
    ):
        """_summary_

        Parameters
        ----------
        variable_type : VariableType, optional
            The variable type we run inference on. One of "continuous", "binary",
            or "counts"; by default "continuous".
        inference_method : InferenceMethod, optional
            The inference method to use. One of "frequentist", "bootstrap",
            or "bayesian"; by default "frequentist"
        hypothesis : Hypothesis, optional
            The directionality of the hypothesis, namely whether the variation
            is "larger", "smaller", or "unequal" compared to the control;
            default "larger". If None provided we use the value configured in
            `spearmint.cfg::hypothesis_test::default_hypothesis`
        alpha : float, optional
            The acceptable Type I error for the inference procedure. If None
            provided, we use the value configured in
            `spearmint.cfg::hypothesis_test::default_alpha`
        metric_name : str, optional
            The name of the metric used to compare groups during the inference
            procedure, mostly used for reporting. By default None
        **inference_procedure_params
            Any additional parameters used to customize a specific inference
            procedure on init.
        """
        self.variable_type = variable_type
        self.inference_method = inference_method
        self.metric_name = metric_name
        self.hypothesis = hypothesis
        self.alpha = alpha

        # Updated after running the inference procedure
        self._results = None

    @abstractmethod
    def _run_inference(
        self, control_samples: Samples, variation_samples: Samples, **inference_kwargs
    ) -> None:
        raise NotImplementedError("Implement me")

    @abstractmethod
    def _make_results(self) -> InferenceResults:
        """
        Generate and return a `InferenceResults` object
        """
        pass

    def run(
        self,
        control_samples: Samples,
        variation_samples: Samples,
    ) -> InferenceResults:
        """
        Run the current inference procedure on control / variation samples.

        Parameters
        ----------
        control : Samples
            the control samples
        variation : Samples
            the variation samples
        **inference_kwargs: kwargs
            Any inference-specific arguments to include in the call signature.

        Returns
        -------
        results : InferenceResults
            The results of running the inference procedure
        """
        self._run_inference(control_samples, variation_samples)
        self._results = self._make_results()  # type: ignore
        return self.results

    @property
    def results(self) -> InferenceResults:
        """
        Raises
        ------
        InferenceResultsMissingError
            if the `run` method hasn't been executed
        """
        if self._results is None:
            raise InferenceResultsMissingError(
                "You likely need to execute the `run` method.",
                " Or check the implementation of `_make_results`",
            )
        return self._results


def get_inference_procedure(
    variable_type: str, inference_method: str, **inference_procedure_params
) -> InferenceProcedure:
    if inference_method == InferenceMethod.frequentist:
        if variable_type == VariableType.continuous:
            from .frequentist.means_delta import MeansDelta as IP

        elif variable_type == VariableType.binary:
            from .frequentist.proportions_delta import ProportionsDelta as IP  # type: ignore

        elif variable_type == VariableType.counts:
            from .frequentist.rates_ratio import RatesRatio as IP  # type: ignore
        else:
            raise ValueError(f"Unknown variable type `{variable_type}`")

    elif inference_method == InferenceMethod.bootstrap:
        from .frequentist.bootstrap_delta import BootstrapDelta as IP  # type: ignore

    elif inference_method == InferenceMethod.bayesian:
        from .bayesian.bayesian_inference import BayesianInferenceProcedure as IP  # type: ignore
    else:
        raise ValueError(f"Unknown inference method `{inference_method}`")

    return IP(
        variable_type=variable_type,
        inference_method=inference_method,
        **inference_procedure_params,
    )



import pytest

from spearmint import Experiment, HypothesisTest
from spearmint.utils import generate_fake_observations


@pytest.fixture()
def counts_data():
    return generate_fake_observations(
        distribution="poisson", n_treatments=6, n_observations=6 * 50, random_seed=123
    )


def test_rates_ratio_larger(counts_data):
    exp = Experiment(data=counts_data)
    ab_test = HypothesisTest(
        metric="metric",
        hypothesis="larger",
        control="A",
        variation="C",
        inference_method="frequentist",
        variable_type="counts",
    )
    results_ab = exp.run_test(ab_test)
    results_ab.display()
    assert results_ab.accept_hypothesis


def test_rates_ratio_smaller(counts_data):
    exp = Experiment(data=counts_data)
    ab_test = HypothesisTest(
        metric="metric",
        hypothesis="smaller",
        control="A",
        variation="C",
        inference_method="frequentist",
        variable_type="counts",
    )
    results_ab = exp.run_test(ab_test)
    results_ab.display()
    assert not results_ab.accept_hypothesis


def test_rates_ratio_unequal(counts_data):
    exp = Experiment(data=counts_data)
    ab_test = HypothesisTest(
        metric="metric",
        hypothesis="unequal",
        control="A",
        variation="C",
        inference_method="frequentist",
        variable_type="counts",
    )
    results_ab = exp.run_test(ab_test)
    results_ab.display()
    assert results_ab.accept_hypothesis


def test_rates_ratio_aa(counts_data):
    exp = Experiment(data=counts_data)
    aa_test = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        inference_method="frequentist",
        variable_type="counts",
    )
    results_aa = exp.run_test(aa_test)
    results_aa.display()
    assert not results_aa.accept_hypothesis


def test_rates_ratio_default(counts_data):
    exp = Experiment(data=counts_data)
    ab_test = HypothesisTest(metric="metric", control="A", variation="C")
    results_ab = exp.run_test(ab_test)
    results_ab.display()
    assert results_ab.accept_hypothesis
    assert ab_test.inference_method == "frequentist"
    assert ab_test.variable_type == "counts"



from dataclasses import dataclass

import numpy as np
import pymc as pm
from arviz import InferenceData

from spearmint.config import N_POSTERIOR_SAMPLES, DEFAULT_PARAMETER_ESTIMATION_METHOD
from spearmint.inference.bayesian.models.analytic_base import BayesianAnalyticModel
from spearmint.inference.inference_base import (
    InferenceProcedure,
    InferenceResults,
    InferenceResultsMissingError,
    OrderedDict,
)
from spearmint.stats import Samples
from spearmint.table import SpearmintTable
from spearmint.typing import Any, Dict, FilePath, Tuple, Union, Optional
from spearmint.utils import format_value, process_warnings

# TODO: Enums for all these
CONTINUOUS_MODEL_NAMES = ["gaussian", "student_t"]

BINARY_MODEL_NAMES = [
    "bernoulli",
    "binomial",
]

COUNTS_MODEL_NAMES = ["poisson"]
SUPPORTED_BAYESIAN_MODEL_NAMES = (
    CONTINUOUS_MODEL_NAMES + BINARY_MODEL_NAMES + COUNTS_MODEL_NAMES
)

DEFAULT_VARIABLE_TYPE_MODELS = {
    "continuous": "gaussian",
    "binary": "binomial",
    "counts": "poisson",
}

SUPPORTED_PARAMETER_ESTIMATION_METHODS = ("mcmc", "advi", "analytic")


class UnsupportedParameterEstimationMethodException(Exception):
    pass


def _get_model_name(
    variable_type: str,
    model_name: Optional[str],
) -> str:
    if model_name is None:
        return DEFAULT_VARIABLE_TYPE_MODELS[variable_type]

    clean_model_name = model_name.replace("-", "_").replace(" ", "_")
    if clean_model_name not in SUPPORTED_BAYESIAN_MODEL_NAMES:
        raise ValueError(f"Unsupported model: {clean_model_name}")
    return clean_model_name


def _get_model_datatype(model_name: str) -> type:
    if model_name in CONTINUOUS_MODEL_NAMES:
        return float
    return int


def _get_delta_param(model_name):
    """Bayesian mnodel param we calculate delta on"""
    if model_name in CONTINUOUS_MODEL_NAMES:
        return "mu"
    elif model_name in BINARY_MODEL_NAMES:
        return "p"
    elif model_name in COUNTS_MODEL_NAMES:
        return "lambda"


class BayesianInferenceResults(InferenceResults):
    """
    Class for storing, displaying, visualizing, and exporting Bayesian
    hypothesis test results.
    """

    def __init__(
        self,
        prior: Samples,
        delta_posterior: Samples,
        delta_relative_posterior: Samples,
        control_posterior: Samples,
        variation_posterior: Samples,
        effect_size_posterior: Samples,
        delta_hdi: Tuple[float, float],
        delta_relative_hdi: Tuple[float, float],
        effect_size_hdi: Tuple[float, float],
        hdi_percentiles: Tuple[float, float],
        prob_greater_than_zero: float,
        prob_less_than_zero: float,
        bayesian_parameter_estimation_method: str,
        model_name: str,
        data_type: type,
        bayesian_model_params: Optional[Dict[str, Any]] = None,
        model_hyperparams: Optional[Dict[str, Any]] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.prior = prior
        self.control_posterior = control_posterior
        self.variation_posterior = variation_posterior
        self.delta_posterior = delta_posterior
        self.delta_relative_posterior = delta_relative_posterior
        self.effect_size_posterior = effect_size_posterior
        self.delta_hdi = delta_hdi
        self.delta_relative_hdi = delta_relative_hdi
        self.effect_size_hdi = effect_size_hdi
        self.prob_greater_than_zero = prob_greater_than_zero
        self.prob_less_than_zero = prob_less_than_zero
        self.bayesian_parameter_estimation_method = bayesian_parameter_estimation_method
        self.data_type = data_type
        self.model_name = model_name
        self.bayesian_model_params = (
            bayesian_model_params if bayesian_model_params else {}
        )
        self.model_hyperparams = model_hyperparams if model_hyperparams else {}
        self.hdi_percentiles = hdi_percentiles

    @property
    def _specific_properties(self):
        """
        Properties specific to the current type of test
        """
        return OrderedDict(
            [
                ("test_type", "bayesian"),
                (
                    f"p({self.variation.name} > {self.control.name})",
                    self.prob_greater_than_zero,
                ),
                ("delta_hdi", self.delta_hdi),
                ("hdi_percentiles", self.hdi_percentiles),
                ("relative_delta_hdi", self.delta_relative_hdi),
                ("effect_size_hdi", self.effect_size_hdi),
                ("credible_mass", 1 - self.alpha),
            ]
        )

    def _render_stats_table(self):
        self._stats_table = BayesianInferenceResultsTable(self)
        return self._stats_table


class BayesianInferenceResultsTable(SpearmintTable):
    def __init__(self, results: BayesianInferenceResults):
        super().__init__(title="Bayesian Delta Results", show_header=False)

        # Add results rows
        self.add_row(
            "Delta",
            format_value(results.delta, precision=4),
        )
        self.add_row(
            "Delta HDI",
            format_value(results.delta_hdi, precision=4),
        )
        self.add_row(
            "Delta Relative",
            format_value(100 * results.delta_relative, precision=2) + " %",
        )
        self.add_row(
            "Delta-relative HDI",
            format_value(100 * np.array(results.delta_relative_hdi), precision=2)
            + " %",
        )
        self.add_row(
            "Effect Size",
            format_value(results.effect_size, precision=4),
        )
        self.add_row(
            "Effect Size HDI",
            format_value(results.effect_size_hdi, precision=4),
        )
        self.add_row(
            "HDI %-tiles",
            format_value(results.hdi_percentiles, precision=4),
        )
        self.add_row(
            "Credible Mass",
            format_value(1 - results.alpha, precision=2),
        )
        self.add_row(
            "Variable Type",
            results.variable_type,
        )
        self.add_row(
            "Inference Method",
            "Bayesian",
        )
        self.add_row(
            "Model Name",
            results.model_name,
        )
        self.add_row(
            "Estimation Method",
            results.bayesian_parameter_estimation_method,
        )
        self.add_row(
            f"p({results.variation.name} > {results.control.name})",
            format_value(results.prob_greater_than_zero, precision=4),
        )
        self.add_row(
            f"p({results.variation.name} < {results.control.name})",
            format_value(results.prob_less_than_zero, precision=4),
        )
        self.add_row(
            "Hypothesis",
            results.hypothesis_text,
        )
        self.add_row(
            "Accept Hypothesis",
            str(results.accept_hypothesis),
        )
        if results.warnings:
            self.add_row(
                "Warnings",
                process_warnings(results.warnings),
            )


def visualize_bayesian_delta_results(
    results: BayesianInferenceResults,
    outfile: Optional[FilePath] = None,
    include_prior: bool = False,
):  # pragma: no cover
    # Lazy import
    import holoviews as hv

    from spearmint import vis

    control_posterior = results.control_posterior
    variation_posterior = results.variation_posterior
    delta_samples = results.delta_posterior
    credible_mass = 1 - results.alpha
    comparison_param = _get_delta_param(results.model_name)

    control_dist = vis.plot_kde(
        samples=control_posterior.data,
        label=results.control.name,
        color=vis.CONTROL_COLOR,
    )

    variation_dist = vis.plot_kde(
        samples=variation_posterior.data,
        label=results.variation.name,
        color=vis.VARIATION_COLOR,
    )

    # Confidence intervals
    control_ci = vis.plot_interval(
        *control_posterior.hdi(credible_mass),
        middle=control_posterior.mean,
        label=results.control.name,
        color=vis.CONTROL_COLOR,
        show_interval_text=True,
    )  # type: ignore # (mypy bug, see #6799)

    variation_ci = vis.plot_interval(
        *variation_posterior.hdi(credible_mass),
        middle=variation_posterior.mean,
        label=results.variation.name,
        color=vis.VARIATION_COLOR,
        show_interval_text=True,
    )  # type: ignore # (mypy bug, see #6799)

    distribution_plot = control_dist * variation_dist * control_ci * variation_ci

    if include_prior:
        prior_dist = vis.plot_kde(
            samples=results.prior.data, label="prior", color=vis.PRIOR_COLOR
        )
        distribution_plot *= prior_dist

    distribution_plot = distribution_plot.relabel(
        f"Posterior {comparison_param} Comparison"
    ).opts(
        legend_position="top_right",
        xlabel=f"Posterior {comparison_param}",
        ylabel="pdf",
    )

    delta_dist = vis.plot_kde(
        samples=delta_samples.data,
        label="Delta Distribution",
        color=vis.DELTA_COLOR,
    )

    max_pdf_height = delta_dist.data["pdf"].max()
    mean_delta = delta_samples.mean

    delta_ci = vis.plot_interval(
        *delta_samples.hdi(credible_mass),
        mean_delta,
        color=vis.NEUTRAL_COLOR,
        label=f"{credible_mass}% HDI",
        show_interval_text=True,
        vertical_offset=-(max_pdf_height * 0.01),
    )  # type: ignore # (mypy bug, see #6799)

    vline = hv.Spikes(([0.0], [max_pdf_height]), vdims="pdf", label="Null Delta").opts(
        color=vis.COLORS.red
    )

    delta_plot = delta_dist * delta_ci * vline
    delta_plot_title = f"Posterior {comparison_param} Delta"
    delta_plot = (
        delta_plot.relabel(delta_plot_title)
        .opts(xlabel="delta", ylabel="pdf")
        .opts(legend_position="top_right")
    )

    visualization = distribution_plot + delta_plot
    visualization.opts(shared_axes=False)

    if outfile is not None:
        vis.save_visualization(visualization, outfile)

    return visualization


@dataclass
class _BayesianModel:
    """Container for Bayesian model object and available fitting options"""

    model_name: str
    model_object: Union[pm.Model, BayesianAnalyticModel]
    mcmc_estimation_supported: bool
    advi_estimation_supported: bool
    analytic_estimation_supported: bool
    hyperparams: Dict[str, Any]


def _build_bayesian_inference_model(
    model_name: str,
    control_samples: Samples,
    variation_samples: Samples,
    bayesian_parameter_estimation_method: str,
    **bayesian_model_params,
) -> _BayesianModel:
    mcmc_estimation_supported = True  # MCMC works for all models
    advi_estimation_supported = False  # ADVI not supported for all models
    analytic_estimation_supported = False  # No Analytic solution for all model

    model_type = (
        "analytic" if bayesian_parameter_estimation_method == "analytic" else "pymc"
    )

    if model_name in CONTINUOUS_MODEL_NAMES:
        from .models import continuous

        advi_estimation_supported = True  # ADVI suppored for all continuous models
        analytic_estimation_supported = True if model_name == "gaussian" else False
        model_builder = getattr(continuous, f"build_{model_name}_{model_type}_model")

    if model_name in BINARY_MODEL_NAMES:
        from .models import binary

        advi_estimation_supported = True if model_name == "bernoulli" else False
        analytic_estimation_supported = True
        model_builder = getattr(binary, f"build_{model_name}_{model_type}_model")

    if model_name in COUNTS_MODEL_NAMES:
        from .models import counts

        advi_estimation_supported = False  # Discrete distributions not supported
        analytic_estimation_supported = True
        model_builder = getattr(counts, f"build_{model_name}_{model_type}_model")

    model_object, hyperparams = model_builder(
        control_samples, variation_samples, **bayesian_model_params
    )

    return _BayesianModel(
        model_name=model_name,
        model_object=model_object,
        mcmc_estimation_supported=mcmc_estimation_supported,
        advi_estimation_supported=advi_estimation_supported,
        analytic_estimation_supported=analytic_estimation_supported,
        hyperparams=hyperparams,
    )


def _fit_model_mcmc(model: _BayesianModel, **inference_kwargs) -> InferenceData:
    if model.mcmc_estimation_supported:
        with model.model_object:  # type: ignore  # context manager only on PyMC model_objects
            return pm.sample(**inference_kwargs)
    raise UnsupportedParameterEstimationMethodException(
        f"MCMC not supported for {model.model_name} model"
    )


def _fit_model_advi(model: _BayesianModel, **inference_kwargs) -> InferenceData:
    if model.advi_estimation_supported:
        with model.model_object:  # type: ignore  # context manager only on PyMC model_objects
            mean_field = pm.fit(method="advi", **inference_kwargs)

            return mean_field.sample(N_POSTERIOR_SAMPLES)

    raise UnsupportedParameterEstimationMethodException(
        f"ADVI parameter estimation not supported for {model.model_name} model"
    )


def _fit_model_analytic(model: _BayesianModel, **inference_kwargs) -> InferenceData:
    if model.analytic_estimation_supported:
        return model.model_object.sample(N_POSTERIOR_SAMPLES)

    raise UnsupportedParameterEstimationMethodException(
        f"Analytic parameter estimation not supported for {model.model_name} model"
    )


class BayesianInferenceProcedure(InferenceProcedure):
    """
    Bayesian inference procedure to test for the difference between two
    samples.
    """

    def __init__(
        self,
        bayesian_model_name: Optional[str] = None,
        bayesian_model_params: Optional[dict] = None,
        bayesian_parameter_estimation_method: str = DEFAULT_PARAMETER_ESTIMATION_METHOD,
        *args,
        **kwargs,
    ):
        """
        Parameters
        ----------
        bayesian_model_name: str (optional)
            The name of the inference model to use:
                -   "gaussian"       : Hierarchical Gaussian model (continuous)
                -   "student-t"      : Hierarchical Student's t model (continuous)
                -   "binomial"       : Beta-Binomial hierarchical model (binary)
                -   "bernoulli"      : Beta-Bernoulli hierarchical model (binary)
                -   "poisson"        : Gamma-Poisson hierarichcal model (counts)
        bayesian_model_params: dict (optional)
            Custom Hyperparameter settings for the Bayesian model. Parameters
            will depend on the model.
        bayesian_parameter_estimation_method: str
            The method used estimate the posterior model parameters. One of:
                -   'mcmc' : use Markov Chain Monte Carlo via PyMC. All models,
                    supported, but may not scale well with large datasets.
                -   'advi' : use Autodiff Variational Inference; scales better than
                    'mcmc', but can only accurately estimate smooth, single-mode
                    posteriors.
                -   'analytic' : uses prior-likelihood conjugacy to estimate the
                    parameters using an analytic solution. Scales best, but only
                    supported by `gaussian`, `beta-binomial`, and `gamma-poisson`
                    models.
            Default is 'mcmc'.
        *args, **kwargs
            Arguments specific for initializing each type of Bayesian model.
            See bayesian inference README for details on specifications for each
            model.
        """
        super().__init__(*args, **kwargs)
        self.model_name = _get_model_name(self.variable_type, bayesian_model_name)
        self.data_type = _get_model_datatype(self.model_name)
        assert (
            bayesian_parameter_estimation_method
            in SUPPORTED_PARAMETER_ESTIMATION_METHODS
        )
        self.bayesian_parameter_estimation_method = bayesian_parameter_estimation_method
        self.bayesian_model_params = (
            bayesian_model_params if bayesian_model_params else {}
        )
        self.inference_data: Optional[InferenceData] = None

    def _process_samples(self, samples: Samples) -> Samples:
        return Samples(np.array(samples.data, dtype=self.data_type), name=samples.name)

    # @abstractmethod
    def _run_inference(
        self, control_samples: Samples, variation_samples: Samples, **inference_kwargs
    ) -> None:
        """
        Run the inference procedure over the samples

        Parameters
        ----------
        control_samples: Samples
            the samples for the control condition
        variation_samples: Samples
            the samples for the varitation condition
        """
        control_samples = self._process_samples(control_samples)
        variation_samples = self._process_samples(variation_samples)

        _bayesian_model = _build_bayesian_inference_model(
            model_name=self.model_name,
            control_samples=control_samples,
            variation_samples=variation_samples,
            bayesian_parameter_estimation_method=self.bayesian_parameter_estimation_method,
            **self.bayesian_model_params,
        )

        if self.bayesian_parameter_estimation_method == "mcmc":
            self.inference_data = _fit_model_mcmc(_bayesian_model, **inference_kwargs)

        elif self.bayesian_parameter_estimation_method == "advi":
            self.inference_data = _fit_model_advi(_bayesian_model, **inference_kwargs)

        elif self.bayesian_parameter_estimation_method == "analytic":
            self.inference_data = _fit_model_analytic(
                _bayesian_model, **inference_kwargs
            )

        # Set attributes
        comparison_param = _get_delta_param(self.model_name)
        self.control = control_samples
        self.variation = variation_samples
        self.prior = self.posterior_samples("prior")
        self.control_posterior = self.posterior_samples(f"{comparison_param}_control")
        self.variation_posterior = self.posterior_samples(
            f"{comparison_param}_variation"
        )
        self.delta_posterior = self.posterior_samples("delta")
        self.delta_relative_posterior = self.posterior_samples("delta_relative")
        self.effect_size_posterior = self.posterior_samples("effect_size")
        self.model_hyperparams = _bayesian_model.hyperparams

    def posterior_samples(self, parameter_name: str) -> Samples:
        """
        Return samples from the model posterior for the requested parameter

        Parameters
        ----------
        parameter_name : str
            The name of the model parameter to sample from the posterior

        Returns
        -------
        posterior_samples : Samples
            Samples from the inference procedure's Bayesian model posterior
            distribution.

        Raises
        ------
        InferenceResultsMissingError
            If the inference procedure has not been run, there is no posterior
            to sample.
        """
        if self.inference_data is not None:
            return Samples(
                observations=self.inference_data.posterior[  # type: ignore  # posterior attribute not initialized by default
                    parameter_name
                ].values.flatten(),
                name=parameter_name,
            )
        raise InferenceResultsMissingError("You may need to execute the .run() method")

    @property
    def accept_hypothesis(self) -> bool:
        prob_greater_than_zero = self.delta_posterior.prob_greater_than(0.0)
        prob_less_than_zero = self.delta_posterior.prob_less_than(0.0)
        if self.hypothesis == "larger":
            return prob_greater_than_zero >= (1 - self.alpha)  # type: ignore

        elif self.hypothesis == "smaller":
            return prob_less_than_zero >= (1 - self.alpha)  # type: ignore
        return (prob_less_than_zero >= (1 - self.alpha)) or (
            prob_greater_than_zero >= (1 - self.alpha)  # type: ignore
        )

    @property
    def test_stats(self) -> Dict[str, Any]:
        return {
            "model_name": self.model_name,
            "delta": self.delta_posterior.mean,
            "delta_relative": self.delta_relative_posterior.mean,
            "delta_hdi": self.delta_posterior.hdi(1 - self.alpha),
            "delta_relative_hdi": self.delta_relative_posterior.hdi(1 - self.alpha),
            "effect_size": self.effect_size_posterior.mean,
            "effect_size_hdi": self.effect_size_posterior.hdi(1 - self.alpha),
            "hdi_percentiles": (self.alpha / 2, 1 - self.alpha / 2),
            "hypothesis": self.hypothesis,
            "prob_greater_than_zero": self.delta_posterior.prob_greater_than(0.0),
            "prob_less_than_zero": self.delta_posterior.prob_less_than(0.0),
            "accept_hypothesis": self.accept_hypothesis,
        }

    # @abstractmethod
    def _make_results(self) -> BayesianInferenceResults:
        """
        Package up inference results
        """
        test_stats = self.test_stats

        results = BayesianInferenceResults(
            control=self.control,
            variation=self.variation,
            prior=self.prior,
            control_posterior=self.control_posterior,
            variation_posterior=self.variation_posterior,
            delta_posterior=self.delta_posterior,
            delta_relative_posterior=self.delta_relative_posterior,
            effect_size_posterior=self.effect_size_posterior,
            bayesian_parameter_estimation_method=self.bayesian_parameter_estimation_method,
            model_name=self.model_name,
            alpha=self.alpha,
            hypothesis=self.hypothesis,
            inference_method=self.inference_method,
            variable_type=self.variable_type,
            metric_name="posterior_delta",
            data_type=self.data_type,
            bayesian_model_params=self.bayesian_model_params,
            model_hyperparams=self.model_hyperparams,
            delta=test_stats["delta"],
            delta_relative=test_stats["delta_relative"],
            delta_hdi=test_stats["delta_hdi"],
            delta_relative_hdi=test_stats["delta_relative_hdi"],
            effect_size=test_stats["effect_size"],
            effect_size_hdi=test_stats["effect_size_hdi"],
            hdi_percentiles=test_stats["hdi_percentiles"],
            prob_greater_than_zero=test_stats["prob_greater_than_zero"],
            prob_less_than_zero=test_stats["prob_less_than_zero"],
            accept_hypothesis=test_stats["accept_hypothesis"],
            visualization_function=visualize_bayesian_delta_results,
        )

        return results



from .bootstrap_delta import BootstrapDelta
from .frequentist_inference import FrequentistInferenceProcedure
from .means_delta import MeansDelta
from .proportions_delta import ProportionsDelta
from .rates_ratio import RatesRatio


__all__ = [
    "BootstrapDelta",
    "FrequentistInferenceProcedure",
    "MeansDelta",
    "ProportionsDelta",
    "RatesRatio",
]



import numpy as np

from spearmint.stats import ProportionComparison, Samples
from spearmint.typing import FilePath, Tuple, Optional

from .frequentist_inference import (
    FrequentistInferenceProcedure,
    FrequentistInferenceResults,
)


def visualize_proportions_delta_results(
    results: FrequentistInferenceResults, outfile: Optional[FilePath] = None
):  # pragma: no cover
    # Lazy import
    import holoviews as hv

    from spearmint import vis

    # Sample distribution comparison plot
    control_dist = vis.plot_binomial(
        p=results.control.mean,
        n=results.control.nobs,
        label=results.control.name,
        color=vis.CONTROL_COLOR,
        alpha=0.5,
    ).opts(axiswise=True)

    variation_dist = vis.plot_binomial(
        p=results.variation.mean,
        n=results.variation.nobs,
        label=results.variation.name,
        color=vis.VARIATION_COLOR,
        alpha=0.5,
    ).opts(axiswise=True)

    def get_binomial_cis(samples, alpha):
        """Convert proportionality to successful # trials"""
        confidence = 1 - alpha
        cis = np.round(
            np.array(samples.confidence_interval(confidence=confidence)) * samples.nobs
        ).astype(int)

        mean = np.round(samples.mean * samples.nobs).astype(int)
        return cis[0], cis[1], mean

    control_ci = vis.plot_interval(
        *get_binomial_cis(results.control, results.alpha),
        label=results.control.name,
        color=vis.CONTROL_COLOR,
        show_interval_text=True,
    )  # type: ignore # (mypy bug, see #6799)

    variation_ci = vis.plot_interval(
        *get_binomial_cis(results.variation, results.alpha),
        label=results.variation.name,
        color=vis.VARIATION_COLOR,
        show_interval_text=True,
    )  # type: ignore # (mypy bug, see #6799)

    distribution_plot = control_dist * variation_dist * control_ci * variation_ci
    distribution_plot = distribution_plot.relabel(
        "Sample Distribution and\nCentral Tendency Estimates"
    ).opts(legend_position="top_right", xlabel="# Successful Trials", ylabel="pdf")

    # Delta distribution plot
    mean_delta = results.variation.mean - results.control.mean
    std_delta = (
        (results.control.var / results.control.nobs)
        + (results.variation.var / results.control.nobs)
    ) ** 0.5

    delta_dist = vis.plot_gaussian(
        mean=mean_delta,
        std=std_delta,
        label="Delta Distribution",
        color=vis.DELTA_COLOR,
    )

    if results.hypothesis == "larger":
        left_bound = results.delta_confidence_interval[0]
        right_bound = np.inf
    elif results.hypothesis == "smaller":
        right_bound = results.delta_confidence_interval[1]
        left_bound = np.inf
    else:
        left_bound = results.delta_confidence_interval[0]
        right_bound = results.delta_confidence_interval[1]

    max_pdf_height = delta_dist.data["pdf"].max()
    ci_percent = round(100 * (1 - results.alpha))
    delta_ci = vis.plot_interval(
        left_bound,
        right_bound,
        mean_delta,
        color=vis.NEUTRAL_COLOR,
        label=f"{ci_percent}% Confidence Interval",
        show_interval_text=True,
        vertical_offset=-(max_pdf_height * 0.01),
    )

    zero_delta_vline = hv.Spikes(
        ([0.0], [max_pdf_height]), vdims="pdf", label="Null Delta"
    ).opts(color=vis.COLORS.red)

    delta_plot = delta_dist * delta_ci * zero_delta_vline
    delta_plot = (
        delta_plot.relabel("Proportionality Delta")
        .opts(xlabel="delta", ylabel="pdf")
        .opts(legend_position="top_right")
    )

    visualization = distribution_plot + delta_plot
    visualization.opts(shared_axes=False)
    if outfile is not None:
        vis.save_visualization(visualization, outfile)

    return visualization


class ProportionsDelta(FrequentistInferenceProcedure):
    """
    Frequentist inference procedure to test for the difference in two sample
    proportions (i.e. conversion rates). Uses Gaussian approximation for sample
    distributions, and thus, assumes adequate sample sizes (i.e. N > 30). If
    this condition is violated, you may receive warnings during inference.
    """

    def __init__(self, variance_assumption: str = "pooled", *args, **kwargs):
        """
        Parameters
        ----------
        variance_assumption : False or float in (0, 1)
            whether to calculate variance based on sample or use control or
            another variance
        """
        super().__init__(*args, **kwargs)
        self.variance_assumption = variance_assumption

    @property
    def test_stats(self) -> dict:
        return self.comparison.z_test_stats

    @property
    def delta_ci(self) -> Tuple[float, float]:
        """
        Calculate confidence interval percentiles and values.
        """
        variance_assumption = (
            self.variance_assumption
            if self.variance_assumption == "pooled"
            else "unequal"
        )
        return self.comparison.zconfint_diff(
            self.alpha, self.hypothesis_sm, variance_assumption
        )

    # @abstractmethod
    def _run_inference(
        self, control_samples: Samples, variation_samples: Samples, **inference_kwargs
    ) -> None:
        """
        Run the inference procedure over the samples with a selected alpha
        value
        """

        self.comparison = ProportionComparison(
            samples_a=variation_samples,
            samples_b=control_samples,
            alpha=self.alpha,
            hypothesis=self.hypothesis,
        )

    # @abstractmethod
    def _make_results(self) -> FrequentistInferenceResults:
        """
        Package up inference results
        """
        test_stats = self.test_stats
        accept_hypothesis = self.accept_hypothesis(test_stats["statistic_value"])

        return FrequentistInferenceResults(
            control=self.comparison.d2,
            variation=self.comparison.d1,
            metric_name=self.metric_name,
            comparison_type="Proportions Delta",
            delta=self.comparison.delta,
            delta_relative=self.comparison.delta_relative,
            effect_size=self.comparison.effect_size,
            alpha=self.comparison.alpha,
            power=self.comparison.power,
            delta_confidence_interval=self.delta_ci,
            delta_confidence_interval_percentiles=self.delta_ci_percentiles,
            inference_method=self.inference_method,
            variable_type=self.variable_type,
            warnings=self.comparison.warnings,
            test_statistic_name=test_stats["statistic_name"],
            test_statistic_value=test_stats["statistic_value"],
            p_value=test_stats["p_value"],
            degrees_freedom=None,
            hypothesis=self.hypothesis,
            accept_hypothesis=accept_hypothesis,
            visualization_function=visualize_proportions_delta_results,
        )



from collections import OrderedDict

import numpy as np
from scipy.stats import norm

from spearmint.inference.inference_base import InferenceProcedure, InferenceResults
from spearmint.stats import CompareMeans
from spearmint.table import SpearmintTable
from spearmint.typing import FilePath, List, Protocol, Tuple, Union, Optional
from spearmint.utils import format_value, process_warnings


class FrequentistInferenceResults(InferenceResults):
    """
    Class for storing, displaying, visualizing, and exporting frequentist
    hypothesis test results.

    delta_confidence_interval : Tuple[float, float]
        The values and percentiles associated with the lower and upper bounds of
        the confidence interval around delta, based on `alpha`
    delta_confidence_interval_percentiles : Tuple[float, float]
        The pdf percentiles associated with `delta_confidence_interval`
    test_statistic_name : str
        The name of the test statistic used to calculate the p-value for the test
    test_statistic_value : float
        The value of the test statistic used to calculate the p-value
    p_value : float in (0, 1)
        The p-value, based on the test value of `statistic_value`
    power : float in (0, 1)
        The statistical power of the experiment, or the probability of correctly
        rejecting the null hypothesis.
    degrees_freedom : int
        The degrees of freedom for the experiment, if applicable (e.g. t-tests)
    correction_method : str
        The name of the multiple comparison correction method used, if any
    """

    def __init__(
        self,
        delta_confidence_interval: Tuple[float, float],
        delta_confidence_interval_percentiles: Tuple[float, float],
        test_statistic_name: str,
        p_value: float,
        power: float = np.nan,
        degrees_freedom: Optional[int] = None,
        test_statistic_value: float = np.nan,
        correction_method: Optional[str] = None,
        warnings: Optional[Union[str, List[str]]] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.power = power
        self.delta_confidence_interval = delta_confidence_interval
        self.delta_confidence_interval_percents = delta_confidence_interval_percentiles
        self.test_statistic_name = test_statistic_name
        self.test_statistic_value = test_statistic_value
        self.p_value = p_value
        self.degrees_freedom = degrees_freedom
        self.correction_method = correction_method
        self.warnings = warnings  # type: ignore
        self.estimate_relative_delta_confidence_interval()

    def estimate_relative_delta_confidence_interval(self):
        """
        Estimate the confidence on the relative difference using interpolation.
        """
        self.delta_ci_relative = tuple(
            100
            * (
                (np.array(self.delta_confidence_interval) + np.array(self.control.mean))
                / self.control.mean
                - 1.0
            )
        )

    @property
    def _specific_properties(self):
        """
        Properties specific to the current type of test
        """
        return OrderedDict(
            [
                ("test_type", "frequentist"),
                ("p", self.p_value),
                ("p_interpretation", "p-value"),
                ("delta_confidence_interval", self.delta_confidence_interval),
                (
                    "delta_confidence_interval_percentiles",
                    self.delta_confidence_interval_percents,
                ),
                ("relative_delta_confidence_interval", self.delta_ci_relative),
                ("ci_interpretation", "Confidence Interval"),
                ("p_value", self.p_value),
                ("power", self.power),
                ("statistic_name", self.test_statistic_name),
                ("statistic_value", self.test_statistic_value),
                ("degrees_freedom", self.degrees_freedom),
                ("mc_correction", self.correction_method),
            ]
        )

    def _render_stats_table(self):
        self._stats_table = FrequentistInferenceResultsTable(self)
        return self._stats_table

    def visualize(self, outfile: Optional[FilePath] = None, *args, **kwargs):
        return self.visualization_function(self, outfile=outfile)


class FrequentistInferenceResultsTable(SpearmintTable):
    def __init__(self, results: FrequentistInferenceResults):
        super().__init__(title=f"{results.comparison_type} Results", show_header=False)

        # Add results rows
        self.add_row(
            "Delta",
            format_value(results.delta, precision=4),
        )
        self.add_row(
            "Delta CI",
            format_value(results.delta_confidence_interval, precision=4),
        )
        self.add_row(
            "Delta-relative",
            format_value(results.delta_relative, precision=4) + " %",
        )
        self.add_row(
            "Delta-relative CI",
            format_value(results.delta_ci_relative, precision=4) + " %",
        )
        self.add_row(
            "Delta CI %-tiles",
            format_value(results.delta_confidence_interval_percents, precision=4),
        )
        self.add_row(
            "Effect Size",
            format_value(results.effect_size, precision=4),
        )
        alpha_corrected = (
            " (corrected)" if results.correction_method is not None else ""
        )
        self.add_row(
            "alpha" + alpha_corrected,
            format_value(results.alpha, precision=3),
        )
        self.add_row(
            "Power",
            format_value(results.power, precision=3),
        )
        self.add_row(
            "Variable Type",
            results.variable_type,
        )
        self.add_row(
            "Inference Method",
            results.inference_method,
        )
        self.add_row(
            f"Test statistic ({results.test_statistic_name})",
            format_value(results.test_statistic_value, precision=2),
        )
        self.add_row(
            "p-value",
            format_value(results.p_value, precision=4),
        )
        if results.degrees_freedom is not None:
            self.add_row(
                "Degrees of Freedom",
                format_value(results.degrees_freedom, precision=0),
            )
        self.add_row(
            "Hypothesis",
            results.hypothesis_text,
        )
        self.add_row(
            "Accept Hypothesis",
            str(results.accept_hypothesis),
        )
        if results.correction_method is not None:
            self.add_row(
                "MC Correction",
                results.correction_method,
            )
        if results.warnings:
            self.add_row(
                "Warnings",
                process_warnings(results.warnings),
            )


class FrequentistTestStatisticalDistribution(Protocol):
    """A distrubiton that implements scipy.stats API"""

    def ppf(self, *args, **kwargs):
        pass


class FrequentistInferenceProcedure(InferenceProcedure):
    def __init__(
        self,
        test_statistic_distribution: FrequentistTestStatisticalDistribution = norm,
        *args,
        **kwargs,
    ):
        """
        Base class for frequentist hypothesis test inference procedures, including
        tests for Means and Proportion deltas. Uses statsmodels.CompareMeans
        objects under the hood

        Parameters
        ----------
        test_statistic_distribution : FrequentistTestStatisticalDistribution, optional
            The probability distribution that captures the shape of the test
            statistic, by default scipy.stats.norm.
        """
        super().__init__(*args, **kwargs)

        self.test_statistic_distribution = test_statistic_distribution
        self.comparison: CompareMeans = None

    @property
    def control_name(self):
        return self.comparison.d2.name

    @property
    def variation_name(self):
        return self.comparison.d1.name

    @property
    def hypothesis_sm(self):
        """
        Statsmodels-compatible hypothesis string
        """
        return self.hypothesis if self.hypothesis != "unequal" else "two-sided"

    def accept_hypothesis(self, statistic_value: float) -> bool:
        """
        Accept the null hypothesis based on the calculated statistic and statitic
        distribution.
        """
        if self.hypothesis == "larger":
            return statistic_value > self.test_statistic_distribution.ppf(
                1 - self.alpha
            )
        elif self.hypothesis == "smaller":
            return statistic_value < self.test_statistic_distribution.ppf(self.alpha)
        elif self.hypothesis == "unequal":
            return abs(statistic_value) > self.test_statistic_distribution.ppf(
                1 - self.alpha / 2.0
            )
        else:
            raise ValueError(f"Unknown hypothesis: {self.hypothesis}")

    @property
    def delta_ci_percentiles(self) -> Tuple[float, float]:
        """Percentiles for the confidence interval on delta, based on `alpha`"""
        if self.hypothesis == "larger":
            return (self.alpha, np.inf)
        elif self.hypothesis == "smaller":
            return (-np.inf, 1 - self.alpha)
        elif self.hypothesis == "unequal":
            return ((self.alpha / 2.0), 1 - (self.alpha / 2.0))
        else:
            raise ValueError("Unknown hypothesis: {!r}".format(self.hypothesis))

    @property
    def test_statistic_name(self) -> str:
        return self.comparison.test_statistic_name



from spearmint.inference.inference_base import (
    InferenceProcedure,
    InferenceResults,
    get_inference_procedure,
)

__all__ = ["InferenceProcedure", "InferenceResults", "get_inference_procedure"]



import os
import shutil
import tempfile

import numpy as np
import pytest

from spearmint import utils


@pytest.fixture()
def testdir():
    tmpdir = tempfile.mkdtemp()
    yield tmpdir
    shutil.rmtree(tmpdir)


def test_rmdir_mkdir(testdir):
    # assert original directory exists
    assert os.path.isdir(testdir)

    # rmdir
    utils.rmdir(testdir)
    assert not os.path.isdir(testdir)

    # mkdir
    utils.mkdir(testdir)
    assert os.path.isdir(testdir)


def test_process_warnings():
    assert utils.process_warnings("a warning") == "a warning"
    assert utils.process_warnings(["a", "warning"]) == "a; warning"
    assert utils.process_warnings([["a"], "warning"]) == "a; warning"
    assert utils.process_warnings([["a"], ["warning"]]) == "a; warning"


def test_coerce_value():
    assert utils.coerce_value("true") is True
    assert utils.coerce_value("false") is False

    assert utils.coerce_value(np.nan) is None
    assert utils.coerce_value(np.inf) is None
    assert utils.coerce_value(-np.inf) is None

    assert isinstance(utils.coerce_value("1.0"), float)
    assert isinstance(utils.coerce_value("1"), int)
    assert isinstance(utils.coerce_value("a,b"), list)


def test_format_value():
    assert utils.format_value(1) == "1"
    assert utils.format_value(1.0) == "1.0"
    assert utils.format_value(1.23456) == "1.2346"
    assert utils.format_value((1, 2)) == "(1, 2)"
    assert utils.format_value((1.0, 2.0)) == "(1.0, 2.0)"
    assert utils.format_value((1.234, 2.345), precision=2) == "(1.23, 2.35)"


def test_ensure_dataframe():
    test_data = utils.generate_fake_observations(distribution="bernoulli")
    with pytest.raises(ValueError):
        utils.ensure_dataframe(None)
    assert utils.ensure_dataframe(test_data).equals(test_data)

    class DataObj:
        data = test_data

    assert utils.ensure_dataframe(DataObj(), "data").equals(test_data)


def test_infer_variable_type():
    assert utils.infer_variable_type(np.array([True, False])) == "binary"
    assert utils.infer_variable_type(np.array([1, 0, 1])) == "binary"
    assert utils.infer_variable_type(np.array([1.0, 0.0, 1.0])) == "binary"
    assert utils.infer_variable_type(np.array([1, 2, 3])) == "counts"
    assert utils.infer_variable_type(np.array([1.0, 2.0, 3.0])) == "counts"
    assert utils.infer_variable_type(np.array([1.0, 2.1, 3.0])) == "continuous"


def test_set_matplotlib_backend():
    assert utils.set_matplotlib_backend() in ("pdf", "agg")


def test_safe_isnan():
    assert utils.safe_isnan(None) is False
    assert utils.safe_isnan(np.inf) == False  # noqa
    assert utils.safe_isnan(np.nan) == True  # noqa


def test_generate_fake_observations():
    fake_bernoulli = utils.generate_fake_observations(distribution="bernoulli")
    assert fake_bernoulli.dtypes["metric"] == bool

    fake_gaussian = utils.generate_fake_observations(distribution="gaussian")
    assert fake_gaussian.dtypes["metric"] == float

    fake_poisson = utils.generate_fake_observations(distribution="poisson")
    assert fake_poisson.dtypes["metric"] == int

    with pytest.raises(ValueError):
        utils.generate_fake_observations(distribution="unsupported distribution")

    n_attrs = 2
    n_treatments = 4
    fake_data = utils.generate_fake_observations(
        n_treatments=n_treatments, n_attributes=n_attrs
    )
    n_columns = len(fake_data.columns)
    n_attr_columns = n_columns - 3  # exclude ID, metric, or treatment columns
    assert n_attr_columns == n_attrs
    assert len(fake_data["treatment"].unique()) == n_treatments



import numpy as np
import pytest

from spearmint.experiment import Experiment
from spearmint.hypothesis_test import CustomMetric, HypothesisTest, HypothesisTestGroup
from spearmint.utils import generate_fake_observations


@pytest.fixture()
def test_data():
    return generate_fake_observations(
        distribution="gaussian", n_treatments=6, n_observations=6 * 50, random_seed=123
    )


def test_hypothesis_test(test_data):
    exp = Experiment(data=test_data)
    default_test = HypothesisTest(
        metric="metric",
        control="A",
        variation="B",
    )
    # Defaults
    assert default_test.inference_method == "frequentist"  # inference method
    assert default_test.variable_type is None  # variable type not inferred
    assert default_test.hypothesis == "larger"

    # Infer datatype from `metric`
    exp.run_test(default_test)

    assert (
        default_test.variable_type == "continuous"
    )  # datatype inferred during `run_test``

    # Copying
    copy_test = default_test.copy(
        variation="C", inference_method="bootstrap", variable_type="binary"
    )
    assert copy_test.variation == "C"
    assert copy_test.inference_method == "bootstrap"
    assert copy_test.variable_type == "binary"

    custom_test = HypothesisTest(
        metric="metric",
        control="A",
        variation="B",
        variable_type="counts",
        inference_method="frequentist",
    )

    assert custom_test.inference_method == "frequentist"
    assert custom_test.variable_type == "counts"  # explicit variable type

    exp.run_test(custom_test)

    assert (
        custom_test.variable_type == "counts"
    )  # explicit variable type doesn't change

    with pytest.raises(ValueError):
        invalid_inference_method_test = HypothesisTest(
            metric="metric",
            control="A",
            variation="B",
            inference_method="invalid_inference_method",
        )
        exp.run_test(invalid_inference_method_test)

    with pytest.raises(ValueError):
        invalid_variable_type_test = HypothesisTest(
            metric="metric",
            control="A",
            variation="B",
            inference_method="frequentist",
            variable_type="invalid_variable_type",
        )
        exp.run_test(invalid_variable_type_test)


def test_hypothesis_test_group(test_data):
    exp = Experiment(data=test_data)

    # run 'A/A' test (should not reject null)
    test_aa = HypothesisTest(
        metric="metric", control="A", variation="A", hypothesis="larger"
    )
    # run A/B test
    test_ab = HypothesisTest(
        metric="metric", control="A", variation="B", hypothesis="larger"
    )
    correction_method = "b"  # shorthand for Bonferonni
    test_group = HypothesisTestGroup(
        tests=[test_aa, test_ab], correction_method=correction_method
    )
    test_group_results = exp.run_test_group(test_group)

    test_group_results.display()

    assert (
        test_group_results.corrected_results[0].correction_method
        == "bonferroni_correction"
    )
    assert test_group_results.original_results[0].correction_method is None

    # corrected alpha should be smaller
    alpha_orig = test_group_results.original_results[0].alpha
    alpha_corrected = test_group_results.corrected_results[0].alpha

    assert alpha_orig > alpha_corrected
    assert not test_group_results.corrected_results[0].accept_hypothesis
    assert test_group_results.corrected_results[1].accept_hypothesis


def test_custom_metric(test_data):
    exp = Experiment(test_data)

    def custom_metric(row):
        return 4 + np.random.rand() if row["treatment"] != "A" else np.random.rand()

    test_ab = HypothesisTest(
        metric=CustomMetric(custom_metric),
        control="A",
        variation="B",
    )
    results_ab = exp.run_test(test_ab)
    assert test_ab.metric_column == "custom_metric"

    results_ab.display()

    assert results_ab.accept_hypothesis
    assert results_ab.test_statistic_name == "z"
    assert results_ab.accept_hypothesis


def test_filter_segments(test_data):
    unsegmented_test = HypothesisTest(metric="metric", control="A", variation="B")
    unsegmented_data = unsegmented_test.filter_segments(test_data)
    assert unsegmented_data.equals(test_data)

    segmented_test = HypothesisTest(
        metric="metric", control="A", variation="B", segmentation="attr_0 == 'A0b'"
    )
    segmented_data = segmented_test.filter_segments(test_data)
    assert np.all(segmented_data.attr_0 == "A0b")


def test_filter_variation(test_data):
    test = HypothesisTest(
        metric="metric", control="A", variation="B", treatment="treatment"
    )

    control_data = test.filter_variations(test_data, test.control)
    variation_data = test.filter_variations(test_data, test.variation)

    assert np.all(control_data.treatment == test.control)
    assert np.all(variation_data.treatment == test.variation)



import numpy as np
from scipy.stats import norm

from spearmint.stats import RateComparison, Samples
from spearmint.typing import FilePath, Tuple, Optional

from .frequentist_inference import (
    FrequentistInferenceProcedure,
    FrequentistInferenceResults,
)


def visualize_rates_ratio_results(
    results: FrequentistInferenceResults, outfile: Optional[FilePath] = None
):  # pragma: no cover
    # Lazy import
    import holoviews as hv

    from spearmint import vis

    # Sample distribution comparison plot
    control_dist = vis.plot_poisson(
        mu=results.control.mean,
        color=vis.CONTROL_COLOR,
        label=results.control.name,
    )

    variation_dist = vis.plot_poisson(
        mu=results.variation.mean,
        color=vis.VARIATION_COLOR,
        label=results.variation.name,
    )

    # Proportion/conversion rate confidence intervals plot
    control_ci = vis.plot_interval(
        *results.control.confidence_interval(1 - results.alpha),
        middle=results.control.mean,
        label=results.control.name,
        color=vis.CONTROL_COLOR,
        show_interval_text=True,
    )

    variation_ci = vis.plot_interval(
        *results.variation.confidence_interval(1 - results.alpha),
        middle=results.variation.mean,
        label=results.variation.name,
        color=vis.VARIATION_COLOR,
        show_interval_text=True,
    )

    distribution_plot = control_dist * variation_dist * control_ci * variation_ci
    distribution_plot = distribution_plot.relabel(
        "Sample Distribution and\nCentral Tendency Estimates"
    ).opts(legend_position="top_right", xlabel="N Events", ylabel="pdf")

    # Delta distribution plot
    mean_ratio = results.variation.mean / results.control.mean
    std_delta = (
        (results.control.var / results.control.nobs)
        + (results.variation.var / results.control.nobs)
    ) ** 0.5

    delta_dist = vis.plot_gaussian(
        mean=mean_ratio,
        std=std_delta,
        label="Delta Distribution",
        color=vis.DELTA_COLOR,
    )
    if results.hypothesis == "larger":
        left_bound = results.delta_confidence_interval[0]
        right_bound = np.inf
    elif results.hypothesis == "smaller":
        right_bound = results.delta_confidence_interval[1]
        left_bound = np.inf
    else:
        left_bound = results.delta_confidence_interval[0]
        right_bound = results.delta_confidence_interval[1]

    max_pdf_height = delta_dist.data["pdf"].max()
    ci_percent = round(100 * (1 - results.alpha))
    delta_ci = vis.plot_interval(
        left=left_bound,
        right=right_bound,
        middle=mean_ratio,
        color=vis.NEUTRAL_COLOR,
        label=f"{ci_percent}% Confidence Interval",
        show_interval_text=True,
        vertical_offset=-(max_pdf_height * 0.01),
    ).opts(ylabel="", xlabel="rate")

    one_delta_vline = hv.Spikes(
        ([1.0], [max_pdf_height]), vdims="pdf", label="Null Ratio"
    ).opts(color=vis.COLORS.red)

    delta_plot = delta_dist * delta_ci * one_delta_vline
    delta_plot = (
        delta_plot.relabel("Rates Ratio")
        .opts(xlabel="ratio", ylabel="pdf")
        .opts(legend_position="top_right")
    )

    visualization = distribution_plot + delta_plot
    visualization.opts(shared_axes=False)

    if outfile is not None:
        vis.save_visualization(visualization, outfile)

    return visualization


class RatesRatio(FrequentistInferenceProcedure):
    """
    Frequentist inference procedure to test for the difference in two sample
    event rates. Use to compare the number of events measured in a standard
    block of time (e.g. clicks in first minute of each user session).
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def test_stats(self) -> dict:
        return self.comparison.rates_test_stats

    @property
    def delta_ci(self) -> Tuple[float, float]:
        """
        Calculate confidence interval for  rates ratio. Intervals outside of 1
        support the alternative hypothesis.

        Calculation follows Li, Tang, & Wong, 2008, using the MOVER-R method,
        and Rao Score intervals (Altman et al, 2000) for individual rate interval
        estimates (aka "FRSI")

        Returns
        -------
        CIs : list
            [(CI_lower, CI_upper), (CI_lower_percentile, CI_upper_percentile)]

        References
        ----------
        Li H.Q, Tang ML, Wong WK (2008) Confidence intervals for ratio of two
            Poisson rates using the method of variance estimates recovery.
            Biometrical Journal 50 (2008)
        Altman D., Machin D., Bryant TN. et al. (2000) "Statistics with confidence"
            (2nd). BMJ Books: Bristol.
        """

        def rao_score_interval(X, z, t):
            # individual rate interval method 2 (Altman et al., 2000)
            a = X + 0.5 * z**2.0
            b = z * np.sqrt(X + 0.25 * z**2.0)
            return (a - b) / t, (a + b) / t

        if self.hypothesis == "larger":
            z = norm.ppf(1 - self.alpha)
        elif self.hypothesis == "smaller":
            z = norm.ppf(self.alpha)
        elif self.hypothesis == "unequal":
            z = np.abs(norm.ppf(1 - self.alpha / 2.0))

        control = self.comparison.d2
        variation = self.comparison.d1

        X1, t1 = control.data.sum(), control.nobs
        X2, t2 = variation.data.sum(), variation.nobs

        lam_1 = X1 / t1
        lam_2 = X2 / t2

        lam_2_lam_1 = lam_2 * lam_1

        l2, u2 = rao_score_interval(X1, z, t1)
        l1, u1 = rao_score_interval(X2, z, t2)

        # Gu et al, 2008; Eq 3
        L = (
            lam_2_lam_1
            - np.sqrt(
                lam_2_lam_1**2 - l1 * (2 * lam_2 - l1) * (u2 * (2 * lam_1 - u2))
            )
        ) / (u2 * (2 * lam_1 - u2))

        # Gu et al, 2008; Eq 4
        U = (
            lam_2_lam_1
            + np.sqrt(
                lam_2_lam_1**2 - u1 * (2 * lam_2 - u1) * (l2 * (2 * lam_1 - l2))
            )
        ) / (l2 * (2 * lam_1 - l2))

        return (L, U)

    # @abstractmethod
    def _run_inference(
        self, control_samples: Samples, variation_samples: Samples, **inference_kwargs
    ) -> None:
        self.comparison = RateComparison(
            samples_a=variation_samples,
            samples_b=control_samples,
            alpha=self.alpha,
            hypothesis=self.hypothesis,
        )

    # @abstractmethod
    def _make_results(self) -> FrequentistInferenceResults:
        test_stats = self.test_stats
        accept_hypothesis = self.accept_hypothesis(test_stats["statistic_value"])

        return FrequentistInferenceResults(
            control=self.comparison.d2,
            variation=self.comparison.d1,
            metric_name=self.metric_name,
            comparison_type="Rates Ratio",
            delta=self.comparison.delta,
            delta_relative=self.comparison.delta_relative,
            effect_size=self.comparison.effect_size,
            alpha=self.comparison.alpha,
            power=self.comparison.power,
            delta_confidence_interval=self.delta_ci,
            delta_confidence_interval_percentiles=self.delta_ci_percentiles,
            inference_method=self.inference_method,
            variable_type=self.variable_type,
            warnings=self.comparison.warnings,
            test_statistic_name=test_stats["statistic_name"],
            test_statistic_value=test_stats["statistic_value"],
            p_value=test_stats["p_value"],
            degrees_freedom=None,
            hypothesis=self.hypothesis,
            accept_hypothesis=accept_hypothesis,
            visualization_function=visualize_rates_ratio_results,
        )



import holoviews as hv
import numpy as np
from holoviews.element.chart import Curve, Histogram

from spearmint import distributions
from spearmint.config import COLORS, FIGURE_PARAMS, POINTS_PLOT_PARAMS
from spearmint.typing import FilePath, Optional, Union
from spearmint.utils import format_value

DEFAULT_COLOR = COLORS.blue
CONTROL_COLOR = COLORS.blue
VARIATION_COLOR = COLORS.green
DELTA_COLOR = COLORS.black
PRIOR_COLOR = COLORS.light_gray

POSITIVE_COLOR = COLORS.green
NEGATIVE_COLOR = COLORS.red
NEUTRAL_COLOR = COLORS.gray


def plot_interval(
    left: float,
    right: float,
    middle: Optional[float] = None,
    color: str = DEFAULT_COLOR,
    label: str = "Interval",
    show_interval_text: bool = False,
    vertical_offset: float = 0.0,
    fontsize: int = 12,
    xlabel: str = "",
    ylabel: str = "",
    **plot_opts,
) -> Curve:  # pragma: no cover
    """Plot an interval spanning (left, right).

    Parameters
    ----------
    left : float
        The lower bound of the interval.
    right : float
        The upper bound of the interval.
    middle : float, optional
        The central anchor point for the interval. If None provided, we assume
        the middle is the average of `left` and `right`.
    color : str, optional
        The color of the interval and any associated text, by default DEFAULT_COLOR
    label : str, optional
        The legend label, by default "Interval"
    show_interval_text : bool, optional
        Whether to display text information describing the interval, by default
        False.
    vertical_offset : float, optional
        The desired vertical location of the interval in the 2D plot, by
        default 0.0
    fontsize : int, optional
        If `show_interval_text=True`, this allows easy customization of the
        interval text font size, by default 12

    Returns
    -------
    interval_curve: Curve
        A holoviews Curve element that can be combined with other elements to
        form an Overlay

    Raises
    ------
    ValueError
        If left and right are invalid values (e.g. -inf/inf), then we can't
        determine reasonable bounds for the visualizatoin
    """
    if middle in (-np.inf, np.inf) and (
        left in (np.inf, -np.inf) or right in (np.inf, -np.inf)
    ):
        raise ValueError("Too many interval values are inf")

    middle = np.mean((left, right)) if middle is None else middle  # type: ignore

    INFTY_SCALE = 2

    _left = middle - INFTY_SCALE * np.abs(right) if left in (np.inf, -np.inf) else left
    _right = (
        middle + INFTY_SCALE * np.abs(left) if right in (np.inf, -np.inf) else right
    )

    middle_point = hv.Points(data=(middle, vertical_offset), label=label).opts(
        color=color, **POINTS_PLOT_PARAMS
    )

    interval_data = ((_left, _right), (vertical_offset, vertical_offset))
    plot_opts.update(FIGURE_PARAMS)
    interval = (
        hv.Curve(data=interval_data, label=label).opts(
            color=color, xlabel=xlabel, ylabel=ylabel, **plot_opts
        )
        * middle_point
    )
    if show_interval_text:
        annotation_text = f"{format_value(middle, precision=2)}\n{format_value((left, right), precision=2)}"
        annotation = hv.Text(
            middle, vertical_offset, annotation_text, fontsize=fontsize, label=label
        ).opts(color=color)
        interval *= annotation

    return interval


def plot_gaussian(
    mean: float = 0,
    std: float = 1,
    color: str = DEFAULT_COLOR,
    label: str = "Gaussian",
    **plot_kwargs,
) -> Curve:  # pragma: no cover
    """
    Plot a Gaussian distribution with parameters `mean` and `std`.

    Parameters
    ----------
    mean : float, optional
        The location of the distribution, by default 0
    std : float, optional
        The spread of the distribution, by default 1
    color : str, optional
        The color of the distrubution line, by default DEFAULT_COLOR
    label : str, optional
        The legend lable for the distribution, by default "Gaussian"

    Returns
    -------
    distribution_curve: Curve
        A holoviews Curve element that can be combined with other elements to
        form an Overlay
    """
    return distributions.Gaussian(mean=mean, std=std, label=label, color=color).plot(
        **plot_kwargs
    )


def plot_bernoulli(
    p: float = 0.5, color: str = DEFAULT_COLOR, label: str = "Bernoulli", **plot_kwargs
) -> Curve:  # pragma: no cover
    """
    Plot a Bernoulli distribution with proportionality parameter `p`.

    Parameters
    ----------
    p : float, optional
        The probability of a positiv outcome, by default 0.5
    color : str, optional
        _description_, by default DEFAULT_COLOR
    color : str, optional
        The color of the distrubution line, by default DEFAULT_COLOR
    label : str, optional
        The legend lable for the distribution, by default "Bernoulli"

    Returns
    -------
    distribution_curve: Curve
        A holoviews Curve element that can be combined with other elements to
        form an Overlay
    """
    return distributions.Bernoulli(p=p, label=label, color=color).plot(**plot_kwargs)


def plot_binomial(
    n: int = 10,
    p: float = 0.5,
    color: str = DEFAULT_COLOR,
    label: str = "Binomial",
    **plot_kwargs,
) -> Curve:  # pragma: no cover
    """
    Plot a Binomial distribution with proportionality parameter `p` and trials
    parameters `n`

    Parameters
    ----------
    n : float, optional
        The number of trials, by default 10
    p : float, optional
        The rate of a positive outcome on a trial, by default 0.5
    color : str, optional
        The color of the distrubution line, by default DEFAULT_COLOR
    label : str, optional
        The legend lable for the distribution, by default "Binomial"

    Returns
    -------
    distribution_curve: Curve
        A holoviews Curve element that can be combined with other elements to
        form an Overlay
    """
    return distributions.Binomial(n=n, p=p, label=label, color=color).plot(
        **plot_kwargs
    )


def plot_poisson(
    mu: float = 10,
    color: str = DEFAULT_COLOR,
    label: str = "Poisson",
    **plot_kwargs,
) -> Curve:  # pragma: no cover
    """Plot a Poisson distribution with rate/mean parameter `mu`.

    Parameters
    ----------
    mu : float, optional
        The location and spread of the distribution, by default 10
    color : str, optional
        The color of the distrubution line, by default DEFAULT_COLOR
    label : str, optional
        The legend lable for the distribution, by default "Poisson"

    Returns
    -------
    distribution_curve: Curve
        A holoviews Curve element that can be combined with other elements to
        form an Overlay
    """
    return distributions.Poisson(mu=mu, label=label, color=color).plot(**plot_kwargs)


def plot_kde(
    samples: np.ndarray,
    color: str = DEFAULT_COLOR,
    label: str = "KDE",
    **plot_kwargs,
) -> Curve:
    """Plot a Kernel-density estimator of the empriical distribution of `samples`

    Parameters
    ----------
    samples : np.ndarray
        The samples to display
    color : str, optional
        The color of the distrubution line, by default DEFAULT_COLOR
    label : str, optional
        The legend lable for the distribution, by default "Poisson"

    Returns
    -------
    distribution_curve: Curve
        A holoviews Curve element.
    """
    return distributions.Kde(samples=samples, label=label, color=color).plot(
        **plot_kwargs
    )


def plot_histogram(
    samples: np.ndarray,
    label: str = "Histogram",
    color: str = DEFAULT_COLOR,
    bins: Union[int, np.ndarray] = 25,
    alpha: float = 0.5,
) -> Histogram:
    """Plot a Histogram of the empriical distribution of `samples`

    Parameters
    ----------
    samples : np.ndarray
        The samples to display
    label : str, optional
        The legend lable for the distribution, by default "Poisson"
    color : str, optional
        The color of the distrubution line, by default DEFAULT_COLOR
    bins : int | np.ndarray
        Similar args to numpy.histogram.
    alpha : float
        The transparency level of the histogram bars.

    Returns
    -------
    histogram: Histogram
        A holoviews Histogram element.
    """
    hist = np.histogram(samples.data, density=True, bins=bins)
    return hv.Histogram(hist, vdims="pdf", label=label).opts(
        facecolor=color, alpha=alpha
    )


def save_visualization(visualization: hv.Element, outfile: FilePath) -> None:
    """Export visualization to disk

    Parameters
    ----------
    visualization : hv.Element
        A holoviews visiualzation element to save
    outfile : FilePath
        The fullfile location on disk to export the visualization.
    """
    hv.save(visualization, outfile)



import copy
from datetime import datetime

from pandas import DataFrame

from spearmint.config import DEFAULT_ALPHA
from spearmint.dataset import Dataset
from spearmint.hypothesis_test import (
    GroupInferenceResults,
    HypothesisTest,
    HypothesisTestGroup,
    InferenceResults,
)
from spearmint.stats import MultipleComparisonCorrection, Samples
from spearmint.typing import List, Optional


class Experiment:
    """
    Container for data and metadata, used to run various hypothesis tests on
    subsets of the data.
    """

    def __init__(
        self,
        data: DataFrame,
        treatment: Optional[str] = None,
        measures: Optional[List[str]] = None,
        attributes: Optional[List[str]] = None,
        metadata: Optional[List[str]] = None,
    ):
        """
        Parameters
        ----------
        data : DataFrame
            the tabular data to analyze, must have columns that correspond
            with `treatment`, `measures`, `attributes`, and `enrollment` if any
            of those are defined.
        treatment : str
            The column in `data` that identifies the association of each
            enrollment in the experiment with one of the experiment conditions. If
            None provided, the global config is searched to identify potential
            treatments in `data`.
        measures : List[str]
            Columns in `data` that are associated with indicator measurements.
            If None provided, the global config is searched to identify potential
            measures in `data`.
        attributes : List[str]
            the columns in `data` that define segmenting attributes
            associated with each enrollment in the experiment. If None provided, the
            global config is searched to identify potential attributes in `data`.
        metadata : List[str]
            Any additional columns in `data` that should be included in the
            experiment dataset. For examples, these columns can be used for
            custom metrics or segmentations.

        name : str
            the name of the experiment
        """

        # metadata
        # if date is None:
        #     self.date = datetime.now().date()
        # elif isinstance(date, datetime):
        #     self.date = date
        # else:
        #     raise ValueError("date parameter must be a datetime.datetime object")

        # self.name = name

        self.created_at = datetime.now()
        self.run_on = None

        # TODO: update `df`` to`data`` in Dataset class for consistency
        self.dataset = Dataset(
            df=data,
            treatment=treatment,
            measures=measures,
            attributes=attributes,
            metadata=metadata,
        )

    @property
    def ds(self):
        """
        Shorthand for the experiment dataset instance
        """
        return self.dataset

    @property
    def measures(self):
        return self.dataset.measures

    @property
    def attributes(self):
        return self.dataset.attributes

    @property
    def observations(self):
        return self.ds.data

    def run_test(
        self,
        test: HypothesisTest,
        alpha: float = DEFAULT_ALPHA,
        correction_method: Optional[str] = None,
        display_results: Optional[bool] = False,
        visualize_results: Optional[bool] = False,
    ) -> InferenceResults:
        """
        Given a HypothesisTest, run the test and return the results.

        Parameters
        ----------
        test : HypothesisTest
            A hypothesis test object
        alpha : float
            The Type I error assumed by the experimenter when running the test
        correction_method: str
            Correction method used when running test (if any)
        display_results : boolean
            Whether to print the test results to stdout
        visualize_results : boolean
            Whether to render a visual representation of the results

        Returns
        -------
        test_results: an instance of a InferenceResults or sublass
            The results of the statistical test.
        """

        control_obs = test.filter_variations(self.dataset, test.control)
        control_obs = test.filter_segments(control_obs)
        control_obs = test.filter_metrics(control_obs)
        control_samples = Samples(control_obs, name=test.control)

        variation_obs = test.filter_variations(self.dataset, test.variation)
        variation_obs = test.filter_segments(variation_obs)
        variation_obs = test.filter_metrics(variation_obs)
        variation_samples = Samples(variation_obs, name=test.variation)

        test_results = test.run(
            control_samples,
            variation_samples,
            alpha=alpha,
            # inference_kwargs=inference_kwargs,
        )
        test_results.correction_method = correction_method

        if display_results:
            test_results.display()

        if visualize_results:
            test_results.visualize()

        test_results.run_at = datetime.now()
        return test_results

    def run_test_group(
        self,
        test_group: HypothesisTestGroup,
        alpha=DEFAULT_ALPHA,
        display_results: bool = False,
        visualize_results: bool = False,
        **inference_kwargs,
    ) -> GroupInferenceResults:
        """
        Run all tests, in a group, performing alpha correction, and adjusting
        the inference results to mitigate multiple comparison error.

        Parameters
        ----------
        test_group : HypothesisTestGroup
            A hypothesis test object
        alpha : float
            The Type I error assumed by the experimenter when running the test
        display_results : boolean
            Whether to print the test results to stdout
        visualize_results : boolean
            Whether to render a visual representation of the results

        Returns
        -------
        test_results: GroupInferenceResults
            The corrected results of the statistical inferences, when applied
            within the context of the group.
        """

        corrected_tests = [copy.deepcopy(t) for t in test_group.tests]

        # run original tests
        original_results = [
            self.run_test(test, alpha, **inference_kwargs) for test in test_group.tests
        ]

        # Run grouped test with correction
        # get p_values for multiple comparison procedure
        p_values = [t.p_value for t in original_results]

        correction_method = test_group.correction_method
        correction = MultipleComparisonCorrection(
            p_values=p_values, alpha=alpha, method=correction_method
        )

        # Rerun hypothesis tests with updated alpha
        corrected_results = [
            self.run_test(
                test,
                alpha=correction.alpha_corrected,
                correction_method=correction.mc_correction_method.__name__,
            )
            for test in corrected_tests
        ]

        test_group_results = GroupInferenceResults(
            test_group=test_group,
            correction=correction,
            original_results=original_results,
            corrected_results=corrected_results,
        )

        if display_results:
            test_group_results.display()

        if visualize_results:
            test_group_results.visualize()

        test_group_results.run_at = datetime.now()

        return test_group_results



from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Protocol, Tuple, Union, Optional

from pandas import DataFrame
from numpy import ndarray

Sequence = Union[List, Tuple, ndarray]
DataFrameColumns = Union[str, List[str]]
FilePath = Union[str, Path]


__all__ = [
    "Path",
    "Any",
    "Callable",
    "Dict",
    "Iterable",
    "List",
    "Protocol",
    "Tuple",
    "Union",
    "DataFrame",
    "DataFrameColumns",
    "FilePath",
    "Sequence",
    "Optional",
]



import json

from spearmint.typing import Iterable, DataFrame, Any, Dict


class JsonableMixin:
    """
    Objects that can be converted to dict and JSON string
    """

    def to_json(self) -> str:
        """Serialize object to JSON string"""
        return json.dumps(self, default=lambda x: x.__dict__, sort_keys=True, indent=4)

    def to_dict(self) -> Dict[Any, Any]:
        """Deserialize object to dictionary"""
        _dict = json.loads(self.to_json())
        assert isinstance(_dict, dict), "to_json must return a dict-able string"
        return _dict

    def __repr__(self) -> str:
        return self.to_json()


class DataframeableMixin:
    """
    Export an object to tabular format. Must overload/implement a `to_dict`
    method that returns a dictionary representation of the object.
    """

    def to_dict(self) -> dict:
        """
        Returns
        -------
        dictionary : dict
            Representation of the object as a dictionary, e.g.
            {
                "ATTR_1_NAME": [ATTR_1_VALUE],
                "ATTR_2_NAME": [[ATTR_2_VALUE_A, FIELD_2_VALUE_B]],
                "ATTR_3_NAME": [(ATTR_3_VALUE_A, FIELD_3_VALUE_B)], ...
            }

        Raises
        ------
        NotImplementedError unless implemented
        """
        raise NotImplementedError()

    def to_dataframe(self) -> DataFrame:
        """
        Export object to a dataframe. If `safe_cast` is True attempt to
        casting common problematic values (see TYPE_MAPPING); this can be useful
        for ensuring type safety, for example when uploading data to a database.
        """

        def is_valid_value(v):
            if v is None:
                return False
            if isinstance(v, Iterable) and len(v) == 0:
                return False
            return True

        _dict = {k: [v] for k, v in self.to_dict().items() if is_valid_value(v)}
        return DataFrame(_dict)

        # for k, v in self.to_dict().items():
        #     # make values a sequence for dataframe API
        #     if v:
        #         _dict[k] = v

        # return DataFrame(self.to_dict())



import pytest

from spearmint import Experiment, HypothesisTest
from spearmint.utils import generate_fake_observations


@pytest.fixture()
def continuous_data():
    return generate_fake_observations(
        distribution="gaussian", n_treatments=6, n_observations=6 * 50, random_seed=123
    )


def test_means_delta_experiment_t(continuous_data):
    """Small sample sizes defautl to t-tests"""
    exp = Experiment(continuous_data.sample(29))

    test_aa = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        hypothesis="unequal",
        inference_method="frequentist",
        variable_type="continuous",
    )
    results_aa = exp.run_test(test_aa)
    results_aa.display()
    assert results_aa.test_statistic_name == "t"


def test_means_delta_experiment_unequal_ab(continuous_data):
    exp = Experiment(continuous_data)

    test_ab = HypothesisTest(
        metric="metric",
        control="A",
        variation="F",
        hypothesis="unequal",
        inference_method="frequentist",
        variable_type="continuous",
    )
    results_ab = exp.run_test(test_ab)
    results_ab.display()
    assert results_ab.test_statistic_name == "z"
    assert results_ab.accept_hypothesis


def test_means_delta_experiment_larger_ab(continuous_data):
    exp = Experiment(continuous_data)

    test_ab = HypothesisTest(
        metric="metric",
        control="A",
        variation="F",
        hypothesis="larger",
        inference_method="frequentist",
        variable_type="continuous",
    )
    results_ab = exp.run_test(test_ab)
    results_ab.display()
    assert results_ab.test_statistic_name == "z"
    assert results_ab.accept_hypothesis


def test_means_delta_experiment_smaller_ab(continuous_data):
    exp = Experiment(continuous_data)

    test_ab = HypothesisTest(
        metric="metric",
        control="A",
        variation="F",
        hypothesis="smaller",
        inference_method="frequentist",
        variable_type="continuous",
    )
    results_ab = exp.run_test(test_ab)
    results_ab.display()
    assert results_ab.test_statistic_name == "z"
    assert not results_ab.accept_hypothesis


def test_means_delta_experiment_aa(continuous_data):
    exp = Experiment(continuous_data)

    test_ab = HypothesisTest(
        metric="metric",
        control="A",
        variation="A",
        hypothesis="larger",
        inference_method="frequentist",
        variable_type="continuous",
    )
    results_ab = exp.run_test(test_ab)
    results_ab.display()
    assert results_ab.test_statistic_name == "z"
    assert not results_ab.accept_hypothesis


def test_means_delta_default(continuous_data):
    exp = Experiment(continuous_data)

    test_ab = HypothesisTest(metric="metric", control="A", variation="B")
    results_ab = exp.run_test(test_ab)
    results_ab.display()
    assert results_ab.accept_hypothesis
    assert test_ab.inference_method == "frequentist"
    assert test_ab.variable_type == "continuous"
    assert results_ab.accept_hypothesis
